import os
import re
import shutil
from dataclasses import dataclass, field
from typing import Dict, List, Optional

try:  # python 3.8+
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from . import julia_template as jl
from .reaction_rules import ReactionRules


@dataclass
class Text2Model(ReactionRules):
    """
    Build a BioMASS-formatted model based on template.

    **reaction** | **parameters** | **initial conditions**

    Attributes
    ----------
    input_txt : str
        Model description file (*.txt), e.g., 'Kholodenko1999.txt'

    lang : Literal["python", "julia"] (default: 'python')
        Either 'python' or 'julia'.

        - 'python': biomass (https://github.com/biomass-dev/biomass)
        - 'julia': BioMASS.jl (https://github.com/biomass-dev/BioMASS.jl)
    """

    input_txt: str
    lang: Literal["python", "julia"] = "python"
    indentation: str = field(default=4 * " ", init=False)

    def __post_init__(self) -> None:
        if not os.path.isfile(self.input_txt):
            raise FileNotFoundError(f"{self.input_txt} does not exist.")
        if self.lang not in ["python", "julia"]:
            raise ValueError("lang must be either 'python' or 'julia'.")
        self.name: str = os.path.splitext(self.input_txt)[0]

    def _update_parameters(self) -> None:
        """
        Update name2idx/parameters.py
        """
        if self.lang == "python":
            with open(
                os.path.join(
                    os.path.dirname(__file__),
                    "template",
                    "name2idx",
                    "parameters.py",
                ),
                encoding="utf-8",
                mode="r",
            ) as f:
                lines = f.readlines()
            for line_num, line in enumerate(lines):
                if line.startswith("NAMES: List[str] = []"):
                    lines[line_num] = "NAMES: List[str] = [\n"
                    lines[line_num] += (
                        f'{self.indentation}"'
                        + f'",\n{self.indentation}"'.join(self.parameters)
                        + '",\n'
                    )
                    lines[line_num] += "]\n"
            with open(
                os.path.join(
                    f"{self.name}",
                    "name2idx",
                    "parameters.py",
                ),
                encoding="utf-8",
                mode="w",
            ) as f:
                f.writelines(lines)
        else:
            lines = jl.PARAMETERS.splitlines()
            for line_num, line in enumerate(lines):
                if line.startswith("const NAMES = []"):
                    lines[line_num] = "const NAMES = [\n"
                    lines[line_num] += (
                        f'{self.indentation}"'
                        + f'",\n{self.indentation}"'.join(self.parameters)
                        + '",\n'
                    )
                    lines[line_num] += "]\n"
            with open(
                os.path.join(
                    f"{self.name}_jl",
                    "name2idx",
                    "parameters.jl",
                ),
                encoding="utf-8",
                mode="w",
            ) as f:
                f.write("\n".join(lines))

    def _update_species(self) -> None:
        """
        Update name2idx/species.py
        """
        if self.lang == "python":
            with open(
                os.path.join(
                    os.path.dirname(__file__),
                    "template",
                    "name2idx",
                    "species.py",
                ),
                encoding="utf-8",
                mode="r",
            ) as f:
                lines = f.readlines()
            for line_num, line in enumerate(lines):
                if line.startswith("NAMES: List[str] = []"):
                    lines[line_num] = "NAMES: List[str] = [\n"
                    lines[line_num] += (
                        f'{self.indentation}"'
                        + '",\n{}"'.format(self.indentation).join(self.species)
                        + '",\n'
                    )
                    lines[line_num] += "]\n"
            with open(
                os.path.join(
                    f"{self.name}",
                    "name2idx",
                    "species.py",
                ),
                encoding="utf-8",
                mode="w",
            ) as f:
                f.writelines(lines)
        else:
            lines = jl.SPECIES.splitlines()
            for line_num, line in enumerate(lines):
                if line.startswith("const NAMES = []"):
                    lines[line_num] = "const NAMES = [\n"
                    lines[line_num] += (
                        '{}"'.format(self.indentation)
                        + '",\n{}"'.format(self.indentation).join(self.species)
                        + '",\n'
                    )
                    lines[line_num] += "]\n"
            with open(
                os.path.join(
                    f"{self.name}_jl",
                    "name2idx",
                    "species.jl",
                ),
                encoding="utf-8",
                mode="w",
            ) as f:
                f.write("\n".join(lines))

    def _update_set_model(self) -> None:
        """Update set_model.py

        If you don't write info., all parameter values and initial values are
        initialized to 1.0 and 0.0, respectively.
        """
        if self.lang == "python":
            with open(
                os.path.join(os.path.dirname(__file__), "template", "set_model.py"),
                encoding="utf-8",
                mode="r",
            ) as f:
                lines = f.readlines()
            # Check species which should be held fixed during simulation.
            for species in self.fixed_species:
                for i, equation in enumerate(self.differential_equations):
                    if f"dydt[V.{species}] = " in equation:
                        self.differential_equations[i] = equation.replace(
                            f"dydt[V.{species}] = ",
                            f"dydt[V.{species}] = 0 # ",
                        )
            for line_num, line in enumerate(lines):
                if line.startswith(2 * self.indentation + "v = {}\n"):
                    # Write flux vector: v
                    lines[line_num + 1] = (
                        2 * self.indentation
                        + f"\n{2 * self.indentation}".join(self.reactions)
                        + "\n\n"
                    )
                elif line.startswith(2 * self.indentation + "dydt = [0] * V.NUM\n"):
                    # Write right-hand side of the differential equation: dydt
                    lines[line_num + 1] = (
                        2 * self.indentation
                        + f"\n{2 * self.indentation}".join(self.differential_equations)
                        + "\n\n"
                    )
                elif line.startswith(self.indentation + "return x"):
                    if not self.param_info:
                        # Set all parameter values = 1.0 if param_info is not given
                        lines[line_num - 1] = (
                            f"{self.indentation + 'x[C.'}"
                            + f"] = 1.0\n{self.indentation + 'x[C.'}".join(self.parameters)
                            + "] = 1.0\n\n"
                        )
                    else:
                        # Set parameter values using param_info
                        lines[line_num - 1] = (
                            "{}".format(self.indentation)
                            + "\n{}".format(self.indentation).join(self.param_info)
                            + "\n\n"
                        )
                elif line.startswith(self.indentation + "return y0"):
                    if self.init_info:
                        # Set initial values using init_info
                        lines[line_num - 1] = (
                            "{}".format(self.indentation)
                            + "\n{}".format(self.indentation).join(self.init_info)
                            + "\n\n"
                        )
            with open(
                os.path.join(
                    f"{self.name}",
                    "set_model.py",
                ),
                encoding="utf-8",
                mode="w",
            ) as f:
                f.writelines(lines)
        else:
            lines = jl.SET_MODEL.splitlines()
            for line_num, line in enumerate(lines):
                if line.startswith(self.indentation + "v = Dict{Int64,Float64}()"):
                    # Write flux vector: v
                    lines[line_num + 1] = (
                        self.indentation + f"\n{self.indentation}".join(self.reactions) + "\n"
                    )
                    lines[line_num + 1] = (
                        lines[line_num + 1]
                        .replace("x[C.", "p[C.")
                        .replace("y[V.", "u[V.")
                        .replace("**", "^")
                    )
                    # Write right-hand side of the differential equation: dydt
                    lines[line_num + 5] = (
                        self.indentation
                        + f"\n{self.indentation}".join(self.differential_equations)
                        + "\n"
                    ).replace("dydt[V.", "du[V.")
                elif line.startswith(self.indentation + "return p"):
                    if not self.param_info:
                        # Set all parameter values = 1.0 if param_info is not given
                        lines[line_num - 1] = (
                            f"{self.indentation + 'p[C.'}"
                            + f"] = 1.0\n{self.indentation + 'p[C.'}".join(self.parameters)
                            + "] = 1.0\n\n"
                        )
                    else:
                        # Set parameter values using param_info
                        lines[line_num - 1] = (
                            "{}".format(self.indentation)
                            + "\n{}".format(self.indentation).join(self.param_info)
                            + "\n\n"
                        ).replace("x[C.", "p[C.")
                elif line.startswith(self.indentation + "return u0"):
                    if self.init_info:
                        # Set initial values using init_info
                        lines[line_num - 1] = (
                            "{}".format(self.indentation)
                            + "\n{}".format(self.indentation).join(self.init_info)
                            + "\n\n"
                        ).replace("y0[V.", "u0[V.")
            with open(
                os.path.join(
                    f"{self.name}_jl",
                    "set_model.jl",
                ),
                encoding="utf-8",
                mode="w",
            ) as f:
                f.write("\n".join(lines))

    def _update_set_search_param(self) -> None:
        """
        Update set_search_param.py
        """
        if self.lang == "python":
            with open(
                os.path.join(os.path.dirname(__file__), "template", "set_search_param.py"),
                encoding="utf-8",
                mode="r",
            ) as f:
                lines = f.readlines()
            for line_num, line in enumerate(lines):
                if "self.idx_params = []" in line:
                    # Write all parameters in idx_params (except for param_excluded)
                    lines[line_num] = f"{2 * self.indentation}self.idx_params = [\n"
                    lines[line_num] += (
                        "{}".format(3 * self.indentation + "C.")
                        + ",\n{}".format(3 * self.indentation + "C.").join(self.parameters)
                        + ",\n"
                    )
                    lines[line_num] += f"{2 * self.indentation}]\n"
                    for param_name in self.param_excluded:
                        # Comment out parameters in param_excluded
                        lines[line_num] = lines[line_num].replace(
                            f"C.{param_name},", f"# C.{param_name},"
                        )
                elif "# parameter constraints" in line:
                    # Add parameter constraints
                    lines[line_num + 1] = (
                        "{}".format(2 * self.indentation)
                        + "\n{}".format(2 * self.indentation).join(self.param_constraints)
                        + "\n\n"
                    )
            with open(
                os.path.join(f"{self.name}", "set_search_param.py"),
                encoding="utf-8",
                mode="w",
            ) as f:
                f.writelines(lines)
        else:
            lines = jl.SET_SEARCH_PARAM.splitlines()
            for line_num, line in enumerate(lines):
                if "search_idx_params::Vector{Int} = []" in line:
                    # Write all parameters in idx_params (except for param_excluded)
                    lines[line_num] = (
                        f"{self.indentation}search_idx_params" + "::Vector{Int} = [\n"
                    )
                    lines[line_num] += (
                        "{}".format(2 * self.indentation + "C.")
                        + ",\n{}".format(2 * self.indentation + "C.").join(self.parameters)
                        + ",\n"
                    )
                    lines[line_num] += f"{self.indentation}]\n"
                    for param_name in self.param_excluded:
                        # Comment out parameters in param_excluded
                        lines[line_num] = lines[line_num].replace(
                            f"C.{param_name},", f"# C.{param_name},"
                        )
                elif "# parameter constraints" in line:
                    # Add parameter constraints
                    lines[line_num + 1] = (
                        "{}".format(self.indentation)
                        + "\n{}".format(self.indentation).join(self.param_constraints)
                        + "\n"
                    ).replace("x[C.", "p[C.")
            with open(
                os.path.join(f"{self.name}_jl", "set_search_param.jl"),
                encoding="utf-8",
                mode="w",
            ) as f:
                f.write("\n".join(lines))

    def _convert_names(
        self,
        line: str,
        p: List[str] = [],
        u: List[str] = [],
        init: List[str] = [],
    ) -> str:
        """
        Replace
        - p[xxx] with x[C.xxx]
        - u[xxx] with sol.y[V.xxx]
        - init[xxx] with y0[V.xxx]

        Parameters
        ----------
        p : list of strings (default: [])
            Parameters.

        u : list of strings (default: [])
            Species.

        init : list of strings (default: [])
            Initial conditions.

        Returns
        -------
        line : string
            Each line.

        """
        for p_name in p:
            if not p_name.strip() in self.parameters:
                raise NameError(f"{p_name.strip()} is not defined in model parameters.")
            else:
                line = (
                    line.replace(f"p[{p_name.strip()}]", f"x[C.{p_name.strip()}]")
                    if self.lang == "python"
                    else line.replace(f"p[{p_name.strip()}]", f"p[C.{p_name.strip()}]")
                )
        for s_name in u:
            if not s_name.strip() in self.species:
                raise NameError(f"{s_name.strip()} is not defined in model species.")
            else:
                line = (
                    line.replace(f"u[{s_name.strip()}]", f"sol.y[V.{s_name.strip()}]")
                    if self.lang == "python"
                    else line.replace(f"u[{s_name.strip()}]", f"sol.u[j][V.{s_name.strip()}]")
                )
        for s_name in init:
            if not s_name.strip() in self.species:
                raise NameError(f"{s_name.strip()} is not defined in model species.")
            else:
                line = (
                    line.replace(f"init[{s_name.strip()}]", f"y0[V.{s_name.strip()}]")
                    if self.lang == "python"
                    else line.replace(f"init[{s_name.strip()}]", f"u0[V.{s_name.strip()}]")
                )
        return line

    def _update_observable(self) -> None:
        """
        Update observable.py
        """
        if self.lang == "python":
            with open(
                os.path.join(os.path.dirname(__file__), "template", "observable.py"),
                encoding="utf-8",
                mode="r",
            ) as f:
                lines = f.readlines()
            if not self.sim_conditions:
                # When sim condition is not given, add 'sim condition control: pass'
                self.sim_conditions = [["control", "pass"]]
            for line_num, line in enumerate(lines):
                if line.startswith(f"{2 * self.indentation}self.obs_names: list = []"):
                    # Write observables
                    lines[line_num] = f"{2 * self.indentation}self.obs_names: list = [\n"
                    lines[line_num] += (
                        f"{3 * self.indentation}'"
                        + f"',\n{3 * self.indentation}'".join(
                            [desc[0].strip() for desc in self.obs_desc]
                        )
                        + "',\n"
                    )
                    lines[line_num] += f"{2 * self.indentation}]\n"
                elif f"self.t: range = range(101)" in line:
                    # Write interval of integration
                    if self.sim_tspan:
                        lines[line_num] = "{}self.t: range = range({}, {}+1)\n".format(
                            2 * self.indentation,
                            self.sim_tspan[0].strip(),
                            self.sim_tspan[1].strip(),
                        )
                elif line.startswith(f"{2 * self.indentation}self.conditions: list = []"):
                    # Write sim.condition
                    lines[line_num] = f"{2 * self.indentation}self.conditions: list = [\n"
                    lines[line_num] += (
                        f"{3 * self.indentation}'"
                        + f"',\n{3 * self.indentation}'".join(
                            [condition[0].strip() for condition in self.sim_conditions]
                        )
                        + "',\n"
                    )
                    lines[line_num] += f"{2 * self.indentation}]\n"
                elif "# unperturbed steady state" in line:
                    if self.sim_unperturbed:
                        lines[line_num + 1] = (
                            2 * self.indentation
                            + f"\n{2 * self.indentation}".join(
                                c.strip() for c in self.sim_unperturbed.split(sep=";")
                            )
                            + "\n"
                        )
                        lines[line_num + 1] += (
                            2 * self.indentation
                            + "y0 = self._get_steady_state(self.diffeq, y0, tuple(x))\n"
                            + 2 * self.indentation
                            + f"if not y0:\n{3 * self.indentation}return False\n"
                        )
                        # pa: parameters
                        # init: initial conditions
                        lines[line_num + 1] = self._convert_names(
                            line=lines[line_num + 1],
                            p=re.findall(r"p\[(.*?)\]", self.sim_unperturbed),
                            init=re.findall(r"init\[(.*?)\]", self.sim_unperturbed),
                        )
                elif "for i, condition in enumerate(self.conditions):" in line:
                    for i, condition in enumerate(self.sim_conditions):
                        # Use ';' for adding description of each condition
                        if i == 0:
                            lines[line_num + 1] = (
                                3 * self.indentation
                                + f'if condition == \'{condition[0].strip(" ")}\':\n'
                                + 4 * self.indentation
                                + f"\n{4 * self.indentation}".join(
                                    c.strip(" ") for c in condition[1].split(sep=";")
                                )
                                + ("\n\n" if len(self.sim_conditions) == 1 else "")
                            )
                        else:
                            lines[line_num + 1] += (
                                3 * self.indentation
                                + f'elif condition == \'{condition[0].strip(" ")}\':\n'
                                + 4 * self.indentation
                                + f"\n{4 * self.indentation}".join(
                                    c.strip(" ") for c in condition[1].split(sep=";")
                                )
                                + ("\n\n" if i == len(self.sim_conditions) - 1 else "")
                            )
                        # pa: parameters
                        # init: initial conditions
                        lines[line_num + 1] = self._convert_names(
                            line=lines[line_num + 1],
                            p=re.findall(r"p\[(.*?)\]", condition[1]),
                            init=re.findall(r"init\[(.*?)\]", condition[1]),
                        )
                elif "sol is None:" in line:
                    lines[line_num + 3] = ""  # initialization (default: pass)
                    for desc in self.obs_desc:
                        lines[line_num + 3] += (
                            "{}".format(4 * self.indentation)
                            + "self.simulations"
                            + f"[self.obs_names.index('{desc[0].strip()}'), :, i] = (\n"
                            + f"{5 * self.indentation}"
                            + desc[1].strip(" ").strip()
                        )
                        # p: parameters
                        # u: species
                        lines[line_num + 3] = self._convert_names(
                            line=lines[line_num + 3],
                            p=re.findall(r"p\[(.*?)\]", desc[1]),
                            u=re.findall(r"u\[(.*?)\]", desc[1]),
                        )
                        lines[line_num + 3] += "\n{}".format(4 * self.indentation + ")\n")
            with open(
                os.path.join(f"{self.name}", "observable.py"),
                encoding="utf-8",
                mode="w",
            ) as f:
                f.writelines(lines)
        else:
            self._split_observables_julia()

    def _split_observables_julia(self) -> None:
        """
        Create observable.jl, simulation.jl and experimental_data.jl
        """
        # observable.jl
        lines = jl.OBSERVABLE.splitlines()
        for line_num, line in enumerate(lines):
            if "const observables = []" in line:
                lines[line_num] = "const observables = [\n"
                lines[line_num + 1] = (
                    f'{self.indentation}"'
                    + f'",\n{self.indentation}"'.join([desc[0].strip() for desc in self.obs_desc])
                    + '",\n'
                )
                lines[line_num + 1] += "]\n"
        with open(
            os.path.join(f"{self.name}_jl", "observable.jl"),
            encoding="utf-8",
            mode="w",
        ) as f:
            f.write("\n".join(lines))
        # simulation.jl
        lines = jl.SIMULATION.splitlines()
        for line_num, line in enumerate(lines):
            if line.startswith("const t = collect(0.0:dt:100.0)"):
                # Write interval of integration
                if self.sim_tspan:
                    lines[line_num] = "const t = collect({}:dt:{})".format(
                        self.sim_tspan[0].strip(),
                        self.sim_tspan[1].strip(),
                    )
            elif line.startswith("const conditions = []"):
                # Write sim.condition
                lines[line_num] = "const conditions = ["
                lines[line_num + 1] = (
                    f'{self.indentation}"'
                    + f'",\n{self.indentation}"'.join(
                        [condition[0].strip() for condition in self.sim_conditions]
                    )
                    + '",\n'
                )
                lines[line_num + 1] += "]\n"
            elif "# unperturbed steady state" in line:
                if self.sim_unperturbed:
                    lines[line_num + 1] = (
                        self.indentation
                        + f"\n{self.indentation}".join(
                            c.strip() for c in self.sim_unperturbed.split(sep=";")
                        )
                        + "\n"
                    )
                    lines[line_num + 1] += (
                        f"{self.indentation}u0 = get_steady_state(diffeq!, u0, p)\n"
                        + f"{self.indentation}if isempty(u0)\n"
                        + f"{2 * self.indentation}return false\n"
                        + f"{self.indentation}end\n"
                    )
                    # p: parameters
                    # init: initial conditions
                    lines[line_num + 1] = self._convert_names(
                        line=lines[line_num + 1],
                        p=re.findall(r"p\[(.*?)\]", self.sim_unperturbed),
                        init=re.findall(r"init\[(.*?)\]", self.sim_unperturbed),
                    )
            elif "for (i, condition) in enumerate(conditions)" in line:
                for i, condition in enumerate(self.sim_conditions):
                    # Use ';' for adding description of each condition
                    if i == 0:
                        lines[line_num + 1] = (
                            3 * self.indentation
                            + f'if condition == "{condition[0].strip(" ")}"\n'
                            + 4 * self.indentation
                            + f"\n{4 * self.indentation}".join(
                                c.strip(" ") for c in condition[1].split(sep=";")
                            )
                            + ("\n\n" if len(self.sim_conditions) == 1 else "")
                        )
                    else:
                        lines[line_num + 1] += (
                            3 * self.indentation
                            + f'elseif condition == "{condition[0].strip(" ")}"\n'
                            + 4 * self.indentation
                            + f"\n{4 * self.indentation}".join(
                                c.strip(" ") for c in condition[1].split(sep=";")
                            )
                            + f"\n{3 * self.indentation}end\n"
                        )
                    # p: parameters
                    # init: initial conditions
                    lines[line_num + 1] = self._convert_names(
                        line=lines[line_num + 1],
                        p=re.findall(r"p\[(.*?)\]", condition[1]),
                        init=re.findall(r"init\[(.*?)\]", condition[1]),
                    )
            elif "if sol === nothing" in line:
                lines[line_num + 4] = ""  # initialization (default: # line_num + 4)
                for desc in self.obs_desc:
                    lines[line_num + 4] += (
                        "{}".format(4 * self.indentation)
                        + "simulations"
                        + f'[observables_index("{desc[0].strip()}"), j, i] = (\n'
                        + f"{5 * self.indentation}"
                        + desc[1].strip(" ").strip()
                    )
                    # p: parameters
                    # u: species
                    lines[line_num + 4] = self._convert_names(
                        line=lines[line_num + 4],
                        p=re.findall(r"p\[(.*?)\]", desc[1]),
                        u=re.findall(r"u\[(.*?)\]", desc[1]),
                    )
                    lines[line_num + 4] += "\n{}".format(4 * self.indentation + ")\n")
        with open(
            os.path.join(f"{self.name}_jl", "simulation.jl"),
            encoding="utf-8",
            mode="w",
        ) as f:
            f.write("\n".join(lines))
        # experimental_data.jl
        lines = jl.EXPERIMENTAL_DATA.splitlines()
        with open(
            os.path.join(f"{self.name}_jl", "experimental_data.jl"),
            encoding="utf-8",
            mode="w",
        ) as f:
            f.write("\n".join(lines))

    def convert(
        self,
        *,
        show_restrictions: bool = False,
        overwrite: bool = False,
    ) -> None:
        """
        Convert text to a biomass-formatted model.

        Parameters
        ----------
        show_restrictions : bool (defauld: :obj:`False`)
            Whether to display reaction indices in which thermodynamic restrictions should be
            imposed. These detailed balance constraints require the product of the equilibrium
            constants along a cycle to be equal to 1.
        overwrite : bool (defauld: :obj:`False`)
            If :obj:`True`, the model folder will be overwritten.

        Examples
        --------
        >>> from pasmopy import Text2Model
        >>> Text2Model("Kholodenko1999.txt").convert()

        """
        if overwrite and os.path.isdir(
            f"{self.name}" if self.lang == "python" else f"{self.name}_jl"
        ):
            shutil.rmtree(f"{self.name}" if self.lang == "python" else f"{self.name}_jl")
        if self.lang == "python":
            shutil.copytree(
                os.path.join(
                    os.path.dirname(__file__),
                    "template",
                ),
                f"{self.name}",
            )
        else:
            os.makedirs(os.path.join(f"{self.name}_jl", "name2idx"))

        self.create_ode()
        self.find_cyclic_reaction_routes()
        self._update_parameters()
        self._update_species()
        self._update_set_model()
        self._update_set_search_param()
        self._update_observable()
        if self.lang == "julia":
            # Create fitness.jl
            lines = jl.FITNESS.splitlines()
            with open(
                os.path.join(f"{self.name}_jl", "fitness.jl"),
                encoding="utf-8",
                mode="w",
            ) as f:
                f.write("\n".join(lines))
        print("Model information\n-----------------")
        print(f"{len(self.reactions):d} reactions")
        print(f"{len(self.species):d} species")
        print(f"{len(self.parameters):d} parameters")
        if show_restrictions:
            if len(self.restrictions) == 0:
                print("No cyclic reaction routes.")
            else:
                print("\nThermodynamic restrictions")
                print("--------------------------")
                for restriction in self.restrictions:
                    print("{" + ", ".join(restriction) + "}")

    def to_markdown(self, n_reaction: int) -> None:
        """
        Create markdown table describing differential equations.

        Parameters
        ----------
        n_reaction : int
            The number of rate equations in the model.

        Examples
        --------
        >>> from pasmopy import Text2Model
        >>> Text2Model("Kholodenko1999.txt").to_markdown(25)

        """
        os.makedirs(os.path.join("markdown", f"{self.name.split(os.sep)[-1]}"), exist_ok=True)
        self.create_ode()
        with open(self.input_txt, encoding="utf-8") as f:
            lines = f.readlines()
        for num, line in enumerate(lines):
            if n_reaction <= num:
                break
            else:
                rate_equation_formatted = f"{self.reactions[num]}".split("=")[1].strip()
                for p_names in self.parameters:
                    if f" ** x[C.{p_names.strip()}]" in f"{self.reactions[num]}":
                        # Hill coefficients
                        rate_equation_formatted = rate_equation_formatted.replace(
                            f" ** x[C.{p_names.strip()}]",
                            f"<sup>{p_names.strip()}</sup>".replace(
                                f"{num + 1:d}", f"<sub>{num + 1:d}</sub>"
                            ),
                        )
                    elif f"x[C.{p_names.strip()}]" in f"{self.reactions[num]}":
                        # Others
                        rate_equation_formatted = rate_equation_formatted.replace(
                            f"x[C.{p_names.strip()}]",
                            f"{p_names.strip()}".replace(
                                f"{num + 1:d}", f"<sub>{num + 1:d}</sub>"
                            ),
                        )
                lines[num] = (
                    f"|{num + 1:d}|"
                    + line.split("|")[0].rstrip()
                    + f"|{rate_equation_formatted.replace('*', '·').replace('y[V.', '[')}|"
                    + "\n"
                )
        with open(
            os.path.join("markdown", f"{self.name.split(os.sep)[-1]}", "rate_equation.md"),
            encoding="utf-8",
            mode="w",
        ) as f:
            f.writelines(
                [
                    "|No.|Reactions|Rate equations|\n|---|---------|--------------|\n",
                    *lines[:n_reaction],
                ]
            )
        # Differential equation
        differential_equations_formatted = [
            eq.replace("*", "·") for eq in self.differential_equations
        ]
        for i, eq in enumerate(differential_equations_formatted):
            for s_name in self.species:
                if f"dydt[V.{s_name.strip()}]" in eq:
                    eq = eq.replace(
                        f"dydt[V.{s_name.strip()}]",
                        f"|{i + 1:d}|d[{s_name.strip()}]/dt",
                    )
                    break
            for num in range(n_reaction):
                if f"v[{num + 1:d}]" in eq:
                    eq = eq.replace(
                        f"v[{num + 1:d}]",
                        f"_v_ <sub>{num + 1:d}</sub>",
                    )
            differential_equations_formatted[i] = eq + "|\n"
        with open(
            os.path.join("markdown", f"{self.name.split(os.sep)[-1]}", "differential_equation.md"),
            encoding="utf-8",
            mode="w",
        ) as f:
            f.writelines(
                [
                    "|No.|Differential equations|\n|---|----------------------|\n",
                    *differential_equations_formatted,
                ]
            )

    def register_word(
        self,
        terminology: Optional[
            Dict[
                Literal[
                    "dimerize",
                    "bind",
                    "dissociate",
                    "is_phosphorylated",
                    "is_dephosphorylated",
                    "phosphorylate",
                    "dephosphorylate",
                    "transcribe",
                    "is_translated",
                    "synthesize",
                    "is_synthesized",
                    "degrade",
                    "is_degraded",
                    "translocate",
                ],
                List[str],
            ]
        ] = None,
    ) -> None:
        """
        Register user-defined rule word.

        Parameters
        ----------
        terminology : Dict[str, List[str]], optional
            Rule to which users register a new rule word and list of user-defined rule words.

        Examples
        --------
        >>> from pasmopy import Text2Model
        >>> mm_kinetics = Text2Model("michaelis_menten.txt")
        >>> mm_kinetics.register_word({"dissociate": ["releases"]})
        >>> mm_kinetics.convert()

        """
        if terminology is None:
            terminology = {}
        for rxn_rule in terminology.keys():
            assert isinstance(rxn_rule, str)
            if rxn_rule not in self.rule_words.keys():
                raise ValueError(
                    f"{rxn_rule} is not defined in reaction_rules.\n"
                    f"Choose a reaction rule from {', '.join(map(str, self.rule_words.keys()))}"
                )
            for my_word in terminology[rxn_rule]:
                assert isinstance(my_word, str)
                for rule, words in self.rule_words.items():
                    for registered_word in words:
                        if " " + my_word in registered_word and registered_word in " " + my_word:
                            raise NameError(
                                f"Cannot supply '{my_word}' to '{rxn_rule}'. "
                                f"Currently, it is used in the rule: {rule}"
                            )
                self.rule_words[rxn_rule].append(" " + my_word)



"""
Model template for conversion of biomass models into
BioMASS.jl (https://github.com/biomass-dev/BioMASS.jl) format.

BioMASS.jl >= 0.5
"""

PARAMETERS: str = """\
module C

const NAMES = []

for (idx,name) in enumerate(NAMES)
    eval(Meta.parse("const $name = $idx"))
end

const NUM = length(NAMES)

end  # module
"""


SPECIES: str = """\
module V

const NAMES = []

for (idx, name) in enumerate(NAMES)
    eval(Meta.parse("const $name = $idx"))
end

const NUM = length(NAMES)

end  # module
"""


SET_MODEL: str = """\
function diffeq!(du, u, p, t)
    v = Dict{Int64,Float64}()

    for i in 1:V.NUM
        @inbounds du[i] = 0.0
    end

end


function param_values()::Vector{Float64}
    p::Vector{Float64} = ones(C.NUM)

    return p
end


function initial_values()::Vector{Float64}
    u0::Vector{Float64} = zeros(V.NUM)

    return u0
end
"""


OBSERVABLE: str = """\
const observables = []

function observables_index(observable_name::String)::Int
    if !(observable_name in observables)
        error("$observable_name is not defined in observables.")
    end
    return findfirst(isequal(observable_name),observables)
end
"""


SIMULATION: str = """\
module Sim
include("./name2idx/parameters.jl")
include("./name2idx/species.jl")
include("./set_model.jl")
include("./observable.jl")

using .C
using .V

using Sundials
using SteadyStateDiffEq

# Options for ODE solver
const ABSTOL = 1e-8
const RELTOL = 1e-8

normalization = Dict{String,Dict{}}()

const dt = 1.0
const t = collect(0.0:dt:100.0)

const conditions = []

simulations = Array{Float64,3}(
    undef, length(observables), length(t), length(conditions)
)


function solveode(
        f::Function,
        u0::Vector{Float64},
        t::Vector{Float64},
        p::Vector{Float64})::Union{ODESolution{},Nothing}
    local sol::ODESolution{}, is_successful::Bool
    try
        prob = ODEProblem(f, u0, (t[1], t[end]), p)
        sol = solve(
            prob,CVODE_BDF(),
            abstol=ABSTOL,
            reltol=RELTOL,
            saveat=dt,
            dtmin=eps(),
            verbose=false
        )
        is_successful = ifelse(sol.retcode === :Success, true, false)
    catch
        is_successful = false
    finally
        if !is_successful
            GC.gc()
        end
    end
    return is_successful ? sol : nothing
end


function get_steady_state(
        f::Function,
        u0::Vector{Float64},
        p::Vector{Float64})::Vector{Float64}
    local sol::SteadyStateSolution{}, is_successful::Bool
    try
        prob = ODEProblem(f, u0, (0.0, Inf), p)
        prob = SteadyStateProblem(prob)
        sol = solve(
            prob,
            DynamicSS(
                CVODE_BDF();
                abstol=ABSTOL,
                reltol=RELTOL
            ),
            dt=dt,
            dtmin=eps(),
            verbose=false
        )
        is_successful = ifelse(sol.retcode === :Success, true, false)
    catch
        is_successful = false
    finally
        if !is_successful
            GC.gc()
        end
    end
    return is_successful ? sol.u : []
end


function simulate!(p::Vector{Float64}, u0::Vector{Float64})::Union{Bool,Nothing}
    # unperturbed steady state

    # add ligand
    for (i, condition) in enumerate(conditions)

        sol = solveode(diffeq!, u0, t, p)
        if sol === nothing
            return false
        else
            @inbounds @simd for j in eachindex(t)
                # line_num + 4
            end
        end
    end
end
end # module
"""


EXPERIMENTAL_DATA: str = """\
module Exp
include("./observable.jl")

experiments = Array{Dict{String,Array{Float64,1}},1}(undef, length(observables))
error_bars = Array{Dict{String,Array{Float64,1}},1}(undef, length(observables))


function get_timepoint(obs_name::String)::Vector{Float64}
    return []
end
end # module
"""


SET_SEARCH_PARAM: str = """\
# Specify model parameters and/or initial values to optimize
function get_search_index()::Tuple{Array{Int64,1},Array{Int64,1}}
    # parameters
    search_idx_params::Vector{Int} = []

    # initial values
    search_idx_initials::Vector{Int} = []

    return search_idx_params, search_idx_initials
end


function get_search_region()::Matrix{Float64}
    p::Vector{Float64} = param_values()
    u0::Vector{Float64} = initial_values()

    search_idx::Tuple{Array{Int64,1},Array{Int64,1}} = get_search_index()
    search_param::Vector{Float64} = initialize_search_param(search_idx, p, u0)

    search_rgn::Matrix{Float64} = zeros(2, length(p) + length(u0))

    # Default: 0.1 ~ 10x
    for (i, j) in enumerate(search_idx[1])
        search_rgn[1,j] = search_param[i] * 0.1  # lower bound
        search_rgn[2,j] = search_param[i] * 10.0  # upper bound
    end

    # Default: 0.5 ~ 2x
    for (i, j) in enumerate(search_idx[2])
        search_rgn[1,j + length(p)] = search_param[i + length(search_idx[1])] * 0.5  # lower bound
        search_rgn[2,j + length(p)] = search_param[i + length(search_idx[1])] * 2.0  # upper bound
    end

    # search_rgn[:, C.param_name] = [lower_bound, upper_bound]
    # search_rgn[:, V.var_name+length(p)] = [lower_bound, upper_bound]

    search_rgn = convert_scale!(search_rgn, search_idx)

    return search_rgn
end


function update_param(indiv::Vector{Float64})::Tuple{Array{Float64,1},Array{Float64,1}}
    p::Vector{Float64} = param_values()
    u0::Vector{Float64} = initial_values()

    search_idx::Tuple{Array{Int64,1},Array{Int64,1}} = get_search_index()

    for (i, j) in enumerate(search_idx[1])
        @inbounds p[j] = indiv[i]
    end
    for (i, j) in enumerate(search_idx[2])
        @inbounds u0[j] = indiv[i + length(search_idx[1])]
    end

    # parameter constraints

    return p, u0
end


function decode_gene2val(indiv_gene)::Vector{Float64}
    search_rgn::Matrix{Float64} = get_search_region()
    indiv::Vector{Float64} = zeros(length(indiv_gene))

    for i in eachindex(indiv_gene)
        indiv[i] = 10^(
            indiv_gene[i] * (
                search_rgn[2,i] - search_rgn[1,i]
            ) + search_rgn[1,i]
        )
    end

    return round.(indiv, sigdigits=7)
end


function encode_val2gene(indiv::Vector{Float64})
    search_rgn::Matrix{Float64} = get_search_region()
    indiv_gene::Vector{Float64} = zeros(length(indiv))

    for i in eachindex(indiv)
        indiv_gene[i] = (
            log10(indiv[i]) - search_rgn[1,i]
        ) / (
            search_rgn[2,i] - search_rgn[1,i]
        )
    end

    return indiv_gene
end


function encode_bestIndivVal2randGene(
        gene_idx::Int64,
        best_indiv::Vector{Float64},
        p0_bounds::Vector{Float64})::Float64
    search_rgn::Matrix{Float64} = get_search_region()
    rand_gene::Float64 = (
        log10(
            best_indiv[gene_idx] * 10^(
                rand() * log10(p0_bounds[2] / p0_bounds[1]) + log10(p0_bounds[1])
            )
        ) - search_rgn[1,gene_idx]
    ) / (
        search_rgn[2,gene_idx] - search_rgn[1,gene_idx]
    )
    return rand_gene
end


function initialize_search_param(
        search_idx::Tuple{Array{Int64,1},Array{Int64,1}},
        p::Vector{Float64},
        u0::Vector{Float64})::Vector{Float64}
    duplicate::Vector{String} = []
    if length(search_idx[1]) != length(unique(search_idx[1]))
        for idx in findall(
            [count(x -> x == i, search_idx[1]) for i in unique(search_idx[1])] .!= 1
        )
            push!(duplicate, C.NAMES[search_idx[1][idx]])
        end
        error(
            "Duplicate parameters (C.): $duplicate"
        )
    elseif length(search_idx[2]) != length(unique(search_idx[2]))
        for idx in findall(
            [count(x -> x == i, search_idx[2]) for i in unique(search_idx[2])] .!= 1
        )
            push!(duplicate, V.NAMES[search_idx[2][idx]])
        end
        error(
            "Duplicate initial conditions (V.): $duplicate"
        )
    end
    search_param = zeros(
        length(search_idx[1]) + length(search_idx[2])
    )
    for (i, j) in enumerate(search_idx[1])
        @inbounds search_param[i] = p[j]
    end
    for (i, j) in enumerate(search_idx[2])
        @inbounds search_param[i + length(search_idx[1])] = u0[j]
    end

    if any(x -> x == 0.0, search_param)
        msg::String = "search_param must not contain zero."
        for idx in search_idx[1]
            if p[idx] == 0.0
                error(
                    @sprintf(
                        "`C.%s` in search_idx_params: ", C.NAMES[idx]
                    ) * msg
                )
            end
        end
        for idx in search_idx[2]
            if u0[idx] == 0.0
                error(
                    @sprintf(
                        "`V.%s` in search_idx_initials: ", V.NAMES[idx]
                    ) * msg
                )
            end
        end
    end

    return search_param
end


function convert_scale!(
        search_rgn::Matrix{Float64},
        search_idx::Tuple{Array{Int64,1},Array{Int64,1}})::Matrix{Float64}
    for i = 1:size(search_rgn, 2)
        if minimum(search_rgn[:,i]) < 0.0
            msg = "search_rgn[lower_bound,upper_bound] must be positive.\n"
            if i <= C.NUM
                error(@sprintf("`C.%s` ", C.NAMES[i]) * msg)
            else
                error(@sprintf("`V.%s` ", V.NAMES[i - C.NUM]) * msg)
            end
        elseif minimum(search_rgn[:,i]) == 0.0 && maximum(search_rgn[:,i]) != 0.0
            msg = "lower_bound must be larger than 0.\n"
            if i <= C.NUM
                error(@sprintf("`C.%s` ", C.NAMES[i]) * msg)
            else
                error(@sprintf("`V.%s` ", V.NAMES[i - C.NUM]) * msg)
            end
        elseif search_rgn[2,i] - search_rgn[1,i] < 0.0
            msg = "lower_bound must be smaller than upper_bound.\n"
            if i <= C.NUM
                error(@sprintf("`C.%s` ", C.NAMES[i]) * msg)
            else
                error(@sprintf("`V.%s` ", V.NAMES[i - C.NUM]) * msg)
            end
        end
    end

    nonzero_idx::Vector{Int} = []
    for i = 1:size(search_rgn, 2)
        if search_rgn[:,i] != [0.0,0.0]
            push!(nonzero_idx, i)
        end
    end
    difference::Vector{Int} = collect(
        symdiff(
            Set(nonzero_idx),
            Set(append!(search_idx[1], C.NUM .+ search_idx[2]))
        )
    )
    if length(difference) > 0
        for idx in difference
            if idx <= C.NUM
                println(@sprintf("`C.%s`", C.NAMES[Int(idx)]))
            else
                println(@sprintf("`V.%s`", V.NAMES[Int(idx) - C.NUM]))
            end
        end
        error(
            "Set these search_params in both search_idx and search_rgn."
        )
    end

    search_rgn = search_rgn[:,nonzero_idx]

    return log10.(search_rgn)
end
"""

FITNESS: str = """\
# Residual Sum of Squares
function compute_objval_rss(
        sim_data::Vector{Float64},
        exp_data::Vector{Float64})::Float64
    error::Float64 = 0.0
    for i in eachindex(exp_data)
        @inbounds error += (sim_data[i] - exp_data[i])^2
    end
    return error
end


# Cosine similarity
function compute_objval_cos(
        sim_data::Vector{Float64},
        exp_data::Vector{Float64})::Float64
    error::Float64 = 1.0 - dot(sim_data,exp_data)/(norm(sim_data)*norm(exp_data))
    return error
end


function conditions_index(condition_name::String)::Int
    if !(condition_name in Sim.conditions)
        error("$condition_name is not defined in Sim.conditions")
    end
    return findfirst(isequal(condition_name), Sim.conditions)
end


function diff_sim_and_exp(
        sim_matrix::Matrix{Float64},
        exp_dict::Dict{String,Array{Float64,1}},
        exp_timepoint::Vector{Float64},
        conditions::Vector{String};
        sim_norm_max::Float64)::Tuple{Vector{Float64}, Vector{Float64}}
    sim_result::Vector{Float64} = []
    exp_result::Vector{Float64} = []

    for (idx,condition) in enumerate(conditions)
        if condition in keys(exp_dict)
            append!(sim_result,sim_matrix[Int.(exp_timepoint.+1),idx])
            append!(exp_result,exp_dict[condition])
        end
    end

    return (sim_result./sim_norm_max, exp_result)
end


# Define an objective function to be minimized.
function objective(indiv_gene)::Float64
    indiv::Vector{Float64} = decode_gene2val(indiv_gene)

    (p,u0) = update_param(indiv)

    if Sim.simulate!(p,u0) isa Nothing
        error::Vector{Float64} = zeros(length(observables))
        for (i,obs_name) in enumerate(observables)
            if isassigned(Exp.experiments,i)
                if length(Sim.normalization) > 0
                    norm_max::Float64 = (
                        Sim.normalization[obs_name]["timepoint"] !== nothing ? maximum(
                            Sim.simulations[
                                i,
                                Sim.normalization[obs_name]["timepoint"],
                                [
                                    conditions_index(c)
                                    for c in Sim.normalization[obs_name]["condition"]
                                ]
                            ]
                        ) : maximum(
                            Sim.simulations[
                                i,
                                :,
                                [
                                    conditions_index(c)
                                    for c in Sim.normalization[obs_name]["condition"]
                                ]
                            ]
                        )
                    )
                end
                error[i] = compute_objval_rss(
                    diff_sim_and_exp(
                        Sim.simulations[i,:,:],
                        Exp.experiments[i],
                        Exp.get_timepoint(obs_name),
                        Sim.conditions,
                        sim_norm_max = ifelse(
                            length(Sim.normalization) == 0, 1.0, norm_max
                        )
                    )...
                )
            end
        end
        return sum(error) # < 1e12
    else
        return 1e12
    end
end
"""



import os
from dataclasses import dataclass, field
from typing import Dict, List

from biomass.exec_model import ModelObject


@dataclass
class WeightingFactors(object):
    """
    Prepare for adding information about gene expression data to model.

    Attributes
    ----------
    model : :py:class:`biomass.exec_model.ModelObject`
        BioMASS model object.
    gene_expression : dict
        Pairs of proteins and their related genes.
    weighting_factors : list of strings
        List of weighting factors.
    prefix : str (default: "w_")
        Prefix of weighting factors on gene expression levels.
    indentation : str (default: 4 spaces)
        How many spaces as indentation.

    Examples
    --------
    >>> import erbb_network
    >>> model = Model(erbb_network.__package__).create()
    >>> gene_expression = {
    ...     "ErbB1": ["EGFR"],
    ...     "ErbB2": ["ERBB2"],
    ...     "ErbB3": ["ERBB3"],
    ...     "ErbB4": ["ERBB4"],
    ...     "Grb2": ["GRB2"],
    ...     "Shc": ["SHC1", "SHC2", "SHC3", "SHC4"],
    ...     "RasGAP": ["RASA1", "RASA2", "RASA3"],
    ...     "PI3K": ["PIK3CA", "PIK3CB", "PIK3CD", "PIK3CG"],
    ...     "PTEN": ["PTEN"],
    ...     "SOS": ["SOS1", "SOS2"],
    ...     "Gab1": ["GAB1"],
    ...     "RasGDP": ["HRAS", "KRAS", "NRAS"],
    ...     "Raf": ["ARAF", "BRAF", "RAF1"],
    ...     "MEK": ["MAP2K1", "MAP2K2"],
    ...     "ERK": ["MAPK1", "MAPK3"],
    ...     "Akt": ["AKT1", "AKT2"],
    ...     "PTP1B": ["PTPN1"],
    ...     "GSK3b": ["GSK3B"],
    ...     "DUSP": ["DUSP5", "DUSP6", "DUSP7"],
    ...     "cMyc": ["MYC"],
    ... }
    >>> weighting_factors = WeightingFactors(model, gene_expression)
    >>> weighting_factors.add_to_params()
    >>> weighting_factors.set_search_bounds()
    """

    model: ModelObject
    gene_expression: Dict[str, List[str]]
    weighting_factors: List[str] = field(default_factory=list, init=False)
    prefix: str = field(default="w_", init=False)
    indentation: str = field(default=" " * 4, init=False)

    def add_to_params(self) -> None:
        """
        Add weighting factors to model parameters.
        """
        for genes in self.gene_expression.values():
            for gene in genes:
                if self.prefix + gene not in self.model.parameters:
                    self.weighting_factors.append(self.prefix + gene)
        if self.weighting_factors:
            with open(
                os.path.join(self.model.path, "name2idx", "parameters.py"),
                mode="r",
                encoding="utf-8",
            ) as f:
                lines = f.readlines()
            for line_num, line in enumerate(lines):
                if line.startswith("NUM: int"):
                    lines[line_num] = "NAMES.extend(\n"
                    lines[line_num] += (
                        f"{self.indentation}[\n"
                        + f'{2 * self.indentation}"'
                        + f'",\n{2 * self.indentation}"'.join(self.weighting_factors)
                        + f'",\n{self.indentation}]\n'
                    )
                    lines[line_num] += ")\n\nNUM: int = len(NAMES)\n"
            with open(
                os.path.join(self.model.path, "name2idx", "parameters.py"),
                mode="w",
                encoding="utf-8",
            ) as f:
                f.writelines(lines)

    def set_search_bounds(self, lb: float = 0.01, ub: float = 100.0) -> None:
        """
        Set search bounds for weighting factors.

        Parameters
        ----------
        lb : float (default: 0.01)
            Lower bound.
        ub : float (default: 100.0)
            Upper bound.
        """
        if self.weighting_factors:
            search_bounds = [
                f"search_rgn[:, C.{wf}] = [{lb}, {ub}]" for wf in self.weighting_factors
            ]
            with open(
                os.path.join(self.model.path, "set_search_param.py"),
                mode="r",
                encoding="utf-8",
            ) as f:
                lines = f.readlines()
            for line_num, line in enumerate(lines):
                if line.startswith(f"{2 * self.indentation}]") and lines[line_num + 3].startswith(
                    f"{2 * self.indentation}self.idx_initials = "
                ):
                    lines[line_num] = (
                        f"{3 * self.indentation}"
                        + f",\n{3 * self.indentation}".join(
                            map(lambda _s: "C." + _s, self.weighting_factors)
                        )
                        + f",\n{2 * self.indentation}]\n"
                    )
                elif line.startswith(f"{2 * self.indentation}search_rgn = convert_scale("):
                    lines[line_num] = 2 * self.indentation + f"\n{2 * self.indentation}".join(
                        search_bounds
                    )
                    lines[line_num] += f"\n\n{2 * self.indentation}search_rgn = convert_scale(\n"
            with open(
                os.path.join(self.model.path, "set_search_param.py"),
                mode="w",
                encoding="utf-8",
            ) as f:
                f.writelines(lines)



from dataclasses import dataclass, field
from typing import Dict, List, Optional

import pandas as pd


@dataclass
class Individualization(object):
    """
    Individualize a mechanistic model by incorporating gene expression levels.

    Attributes
    ----------
    parameters : List[str]
        List of model parameters.

    species : List[str]
        List of model species.

    transcriptomic_data : str
        Path to normalized gene expression data (CSV-formatted),
        e.g., (1) RLE-normalized and (2) post-ComBat TPM values.
        Below is an example of data table.

        =========== ======== ======== ======== ===
        Description patient1 patient2 patient3 ...
        =========== ======== ======== ======== ===
        gene1       value1,1 value1,2 value1,3 ...
        gene2       value2,1 value2,2 value2,3 ...
        gene3       value3,1 value3,2 value3,3 ...
        ...         ...      ...      ...      ...
        =========== ======== ======== ======== ===

    gene_expression : Dict[str, List[str]]
        Pairs of proteins and their related genes.

    read_csv_kws : dict, optional
        Keyword arguments to pass to ``pandas.read_csv``.

    prefix : str (default: "w_")
        Prefix of weighting factors on gene expression levels.

    Examples
    --------
    ``set_search_param.py``

    .. code-block:: python

        import os
        import numpy as np
        from pasmopy import Individualization
        from . import __path__
        from .name2idx import C, V
        from .set_model import initial_values, param_values

        incorporating_gene_expression_levels = Individualization(
            parameters=C.NAMES,
            species=V.NAMES,
            transcriptomic_data=os.path.join("transcriptomic_data", "TPM_RLE_postComBat_BRCA_BREAST.csv"),
            gene_expression={
                "ErbB1": ["EGFR"],
                "ErbB2": ["ERBB2"],
                "ErbB3": ["ERBB3"],
                "ErbB4": ["ERBB4"],
                "Grb2": ["GRB2"],
                "Shc": ["SHC1", "SHC2", "SHC3", "SHC4"],
                "RasGAP": ["RASA1", "RASA2", "RASA3"],
                "PI3K": ["PIK3CA", "PIK3CB", "PIK3CD", "PIK3CG"],
                "PTEN": ["PTEN"],
                "SOS": ["SOS1", "SOS2"],
                "Gab1": ["GAB1"],
                "RasGDP": ["HRAS", "KRAS", "NRAS"],
                "Raf": ["ARAF", "BRAF", "RAF1"],
                "MEK": ["MAP2K1", "MAP2K2"],
                "ERK": ["MAPK1", "MAPK3"],
                "Akt": ["AKT1", "AKT2"],
                "PTP1B": ["PTPN1"],
                "GSK3b": ["GSK3B"],
                "DUSP": ["DUSP5", "DUSP6", "DUSP7"],
                "cMyc": ["MYC"],
            },
            read_csv_kws={"index_col": "Description"}
        )

        ...

        def update(self, indiv):
            x = param_values()
            y0 = initial_values()
            for i, j in enumerate(self.idx_params):
                x[j] = indiv[i]
            for i, j in enumerate(self.idx_initials):
                y0[j] = indiv[i + len(self.idx_params)]
            # As maximal transcription rate
            x[C.V291] = incorporating_gene_expression_levels.as_reaction_rate(
                __path__[0].split(os.sep)[-1], x, "V291", "DUSP"
            )
            x[C.V310] = incorporating_gene_expression_levels.as_reaction_rate(
                __path__[0].split(os.sep)[-1], x, "V310", "cMyc"
            )
            # As initial conditions
            y0 = incorporating_gene_expression_levels.as_initial_conditions(
                __path__[0].split(os.sep)[-1], x, y0
            )

            ...
    """

    parameters: List[str]
    species: List[str]
    transcriptomic_data: str
    gene_expression: Dict[str, List[str]]
    read_csv_kws: Optional[dict] = field(default=None)
    prefix: str = field(default="w_", init=False)

    def __post_init__(self) -> None:
        kwargs = self.read_csv_kws
        if kwargs is None:
            kwargs = {}
        self._expression_level: pd.DataFrame = pd.read_csv(self.transcriptomic_data, **kwargs)

    @property
    def expression_level(self) -> pd.DataFrame:
        return self._expression_level

    def _calculate_weighted_sum(
        self,
        id: str,
        x: List[float],
    ) -> Dict[str, float]:
        """
        Incorporate gene expression levels in the model.

        Returns
        -------
        weighted_sum : Dict[str, float]
            Estimated protein levels after incorporating transcriptomic data.
        """
        weighted_sum = dict.fromkeys(self.gene_expression, 0.0)
        for (protein, genes) in self.gene_expression.items():
            for gene in genes:
                weighted_sum[protein] += (
                    x[self.parameters.index(self.prefix + gene)]
                    * self.expression_level.at[gene, id]
                )
        return weighted_sum

    def as_reaction_rate(
        self,
        id: str,
        x: List[float],
        param_name: str,
        protein: str,
    ) -> float:
        """
        Gene expression levels are incorporated as a reaction rate.

        Parameters
        ----------
        id : str
            CCLE_ID or TCGA_ID.

        x : List[float]
            List of parameter values.

        param_name : str
            Name of the parameter incorporating gene_expression_data.

        protein: str
            Protein involved in the reaction.

        Returns
        -------
        param_value : float
        """
        weighted_sum = self._calculate_weighted_sum(id, x)
        param_value = x[self.parameters.index(param_name)]
        param_value *= weighted_sum[protein]
        return param_value

    def as_initial_conditions(
        self,
        id: str,
        x: List[float],
        y0: List[float],
    ) -> List[float]:
        """
        Gene expression levels are incorporated as initial conditions.

        Parameters
        ----------
        id : str
            CCLE_ID or TCGA_ID.

        x : List[float]
            List of parameter values.

        y0 : List[float]
            List of initial values.

        Returns
        -------
        y0 (individualized) : List[float]
            Cell-line- or patient-specific initial conditions.
        """
        weighted_sum = self._calculate_weighted_sum(id, x)
        for protein in self.gene_expression.keys():
            y0[self.species.index(protein)] *= weighted_sum[protein]
        return y0



"""Patient-Specific Modeling in Python"""

from biomass.core import *

from .construction import Text2Model
from .individualization import Individualization
from .patient_model import PatientModelAnalyses, PatientModelSimulations
from .version import __version__

__author__ = __maintainer__ = "Hiroaki Imoto"
__email__ = "himoto@protein.osaka-u.ac.jp"



import csv
import multiprocessing
import os
from dataclasses import dataclass, field
from typing import Callable, Dict, List, Optional, Union

try:  # python 3.8+
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import numpy as np
import pandas as pd
import seaborn as sns
from biomass import Model, run_analysis, run_simulation
from scipy.integrate import simpson
from tqdm import tqdm


@dataclass
class InSilico(object):
    """
    Patient-specific in silico models.

    Attributes
    ----------
    path_to_models : str
        Path (dot-separated) to the directory containing patient-specific models.

    patients : list of strings
        List of patients' names or identifiers.
    """

    path_to_models: str
    patients: List[str]

    def __post_init__(self) -> None:
        """
        Check for duplicates in self.patients.
        """
        duplicate = [patient for patient in set(self.patients) if self.patients.count(patient) > 1]
        if duplicate:
            raise NameError(f"Duplicate patient: {', '.join(duplicate)}")

    def parallel_execute(
        self,
        func: Callable[[str], None],
        n_proc: int,
        method: Literal["spawn", "fork", "forkserver"],
    ) -> None:
        """
        Execute multiple models in parallel.

        Parameters
        ----------
        func : Callable
            Function executing a single patient-specific model.

        n_proc : int
            The number of worker processes to use.

        method : Literal["spawn", "fork", "forkserver"]
            Start method in ``multiprocessing``.
        """

        ctx = multiprocessing.get_context(method)
        p = ctx.Pool(processes=n_proc)
        with tqdm(total=len(self.patients)) as t:
            for _ in p.imap_unordered(func, self.patients):
                t.update(1)
        p.close()

    @staticmethod
    def _check_ctx(context: str) -> None:
        """
        Check whether the context is appropriate.
        """
        contexts = ["spawn", "fork", "forkserver"]
        if context not in contexts:
            raise ValueError("context must be one of '{}'.".format("', '".join(contexts)))


@dataclass
class PatientModelSimulations(InSilico):
    """
    Run simulations of patient-specific models.

    Attributes
    ----------
    biomass_kws : dict, optional
        Keyword arguments to pass to ``biomass.run_simulation``.

    response_characteristics : dict[str, Callable[[1d-array], int ot float]]
        A dictionary containing functions to extract dynamic response characteristics
        from time-course simulations.
    """

    biomass_kws: Optional[dict] = field(default=None)
    response_characteristics: Dict[str, Callable[[np.ndarray], Union[int, float]]] = field(
        default_factory=lambda: dict(
            max=np.max,
            AUC=simpson,
        ),
        init=False,
    )

    def _run_single_patient(self, patient: str) -> None:
        """
        Run a single patient-specifc model simulation.
        """

        kwargs = self.biomass_kws
        if kwargs is None:
            kwargs = {}
        kwargs.setdefault("viz_type", "average")
        kwargs.setdefault("stdev", True)

        model = Model(".".join([self.path_to_models, patient.strip()])).create()
        run_simulation(model, **kwargs)

    def run(
        self,
        n_proc: Optional[int] = None,
        context: Literal["spawn", "fork", "forkserver"] = "spawn",
    ) -> None:
        """
        Run simulations of multiple patient-specific models in parallel.

        Parameters
        ----------
        n_proc : int, optional
            The number of worker processes to use.

        context : Literal["spawn", "fork", "forkserver"] (default: "spawn")
            The context used for starting the worker processes.
        """
        if n_proc is None:
            n_proc = multiprocessing.cpu_count() - 1
        self._check_ctx(context)
        self.parallel_execute(self._run_single_patient, n_proc, context)

    @staticmethod
    def _cleanup_csv(dirname: str) -> None:
        """
        Delete CSV files in folder.

        Parameters
        ----------
        dirname: str
            Name of the directory.
        """
        files = os.listdir(dirname)
        for file in files:
            if file.endswith(".csv"):
                os.remove(os.path.join(dirname, f"{file}"))

    def _extract(
        self,
        dynamical_features: Dict[str, Dict[str, List[str]]],
        normalization: bool,
    ) -> None:
        """
        Extract response characteristics from patient-specific signaling dynamics.
        """
        os.makedirs("classification", exist_ok=True)
        self._cleanup_csv("classification")
        for obs_name, conditions_and_metrics in dynamical_features.items():
            with open(
                os.path.join("classification", f"{obs_name}.csv"),
                "w",
                newline="",
            ) as f:
                writer = csv.writer(f)
                header = ["Sample"]
                for condition, metrics in conditions_and_metrics.items():
                    for metric in metrics:
                        header.append(f"{condition}_{metric}")
                writer.writerow(header)

                for patient in tqdm(self.patients):
                    patient_specific = Model(
                        ".".join([self.path_to_models, patient.strip()])
                    ).create()
                    all_data = np.load(
                        os.path.join(
                            patient_specific.path,
                            "simulation_data",
                            "simulations_all.npy",
                        )
                    )
                    data = np.array(all_data[patient_specific.observables.index(obs_name)])
                    if normalization:
                        for i in range(data.shape[0]):
                            if not np.isnan(data[i]).all() and not np.all(data[i] == 0.0):
                                data[i] /= np.nanmax(data[i])
                        data = np.nanmean(data, axis=0)
                        if not np.all(data == 0.0):
                            data /= np.max(data)
                    patient_specific_characteristics = [patient]
                    for h in header[1:]:
                        condition, metric = h.split("_")
                        patient_specific_characteristics.append(
                            str(
                                self.response_characteristics[metric](
                                    data[:, patient_specific.problem.conditions.index(condition)],
                                )
                            )
                        )
                    writer = csv.writer(f, lineterminator="\n")
                    writer.writerow(patient_specific_characteristics)

    def subtyping(
        self,
        fname: Optional[str],
        dynamical_features: Dict[str, Dict[str, List[str]]],
        normalization: bool = True,
        *,
        clustermap_kws: Optional[dict] = None,
    ):
        """
        Classify patients based on dynamic characteristics extracted from simulation results.

        Parameters
        ----------
        fname : str, path-like or :obj:`None`
            The clustermap is saved as fname if it is not :obj:`None`.

        dynamical_features : Dict[str, Dict[str, List[str]]]
            ``{"observable": {"condition": ["metric", ...], ...}, ...}``.
            Characteristics in the signaling dynamics used for classification.

        normalization : bool (default: :obj:`True`)
            Whether to perform max-normalization.

        clustermap_kws : dict, optional
            Keyword arguments to pass to ``seaborn.clustermap()``.

        Examples
        --------
        Subtype classification

        >>> with open("models/breast/sample_names.txt", mode="r") as f:
        ...    TCGA_ID = f.read().splitlines()
        >>> from pasmopy import PatientModelSimulations
        >>> simulations = PatientModelSimulations("models.breast", TCGA_ID)
        >>> simulations.subtyping(
        ...    "subtype_classification.pdf",
        ...    {
        ...        "Phosphorylated_Akt": {"EGF": ["max"], "HRG": ["max"]},
        ...        "Phosphorylated_ERK": {"EGF": ["max"], "HRG": ["max"]},
        ...        "Phosphorylated_c-Myc": {"EGF": ["max"], "HRG": ["max"]},
        ...    },
        ...    clustermap_kws={"figsize": (9, 12)}
        ... )

        Add new characteristics

        >>> import numpy as np
        >>> def get_droprate(time_course: np.ndarray) -> float:
        ...     return - (time_course[-1] - np.max(time_course)) / (len(time_course) - np.argmax(time_course))
        >>> simulations.response_characteristics["droprate"] = get_droprate
        """
        # seaborn clustermap
        if clustermap_kws is None:
            clustermap_kws = {}
        clustermap_kws.setdefault("z_score", 1)
        clustermap_kws.setdefault("cmap", "RdBu_r")
        clustermap_kws.setdefault("center", 0)
        # extract response characteristics
        self._extract(dynamical_features, normalization)
        if fname is not None:
            characteristics: List[pd.DataFrame] = []
            files = os.listdir("classification")
            for file in files:
                observable, ext = os.path.splitext(file)
                if ext == ".csv":
                    df = pd.read_csv(os.path.join("classification", file), index_col="Sample")
                    characteristics.append(
                        df.rename(columns=lambda s: observable.replace("_", " ") + "_" + s)
                    )
            all_info = pd.concat(characteristics, axis=1)
            all_info.index.name = ""
            fig = sns.clustermap(all_info, **clustermap_kws)
            fig.savefig(fname)


@dataclass
class PatientModelAnalyses(InSilico):
    """
    Run analyses of patient-specific models.

    Attributes
    ----------
    biomass_kws : dict, optional
        Keyword arguments to pass to ``biomass.run_analysis``.
    """

    biomass_kws: Optional[dict] = field(default=None)

    def _run_single_patient(self, patient: str) -> None:
        """
        Run a single patient-specifc model analysis.
        """

        kwargs = self.biomass_kws
        if kwargs is None:
            kwargs = {}
        kwargs.setdefault("target", "initial_condition")
        kwargs.setdefault("metric", "integral")
        kwargs.setdefault("style", "heatmap")
        kwargs.setdefault("options", None)

        model = Model(".".join([self.path_to_models, patient.strip()])).create()
        run_analysis(model, **kwargs)

    def run(
        self,
        n_proc: Optional[int] = None,
        context: Literal["spawn", "fork", "forkserver"] = "spawn",
    ) -> None:
        """
        Run analyses of multiple patient-specific models in parallel.

        Parameters
        ----------
        n_proc : int, optional
            The number of worker processes to use.

        context : Literal["spawn", "fork", "forkserver"] (default: "spawn")
            The context used for starting the worker processes.
        """
        if n_proc is None:
            n_proc = multiprocessing.cpu_count() - 1
        self._check_ctx(context)
        self.parallel_execute(self._run_single_patient, n_proc, context)



import sys
import warnings
from dataclasses import dataclass, field
from difflib import SequenceMatcher
from typing import Dict, List, NamedTuple, Optional

import numpy as np

from .thermodynamic_restrictions import ComplexFormation, DuplicateError, ThermodynamicRestrictions


class UnregisteredRule(NamedTuple):
    expected: Optional[str]
    original: Optional[str]


PREPOSITIONS: List[str] = [
    "to",
    "for",
    "from",
    "up",
    "down",
    "in",
    "on",
    "at",
    "off",
    "into",
    "around",
    "among",
    "between",
    "of",
    "over",
    "above",
    "below",
    "under",
    "through",
    "across",
    "along",
    "near",
    "by",
    "beside",
]


@dataclass
class ReactionRules(ThermodynamicRestrictions):
    """Create an executable biochemical model from text.

    .. list-table:: Available reaction rules
        :widths: 25 50 25
        :header-rows: 1

        * - Rule
          - Example sentence
          - Parameters (optional)
        * - :func:`~pasmopy.construction.reaction_rules.dimerize`
          - *A* dimerizes <--> *AA*
          - .. math:: kf, kr
        * - :func:`~pasmopy.construction.reaction_rules.bind`
          - *A* binds *B* <--> *AB*
          - .. math:: kf, kr
        * - :func:`~pasmopy.construction.reaction_rules.dissociate`
          - *AB* dissociates to *A* and *B*
          - .. math:: kf, kr
        * - :func:`~pasmopy.construction.reaction_rules.is_phosphorylated`
          - *uA* is phosphorylated <--> *pA*
          - .. math:: kf, kr
        * - :func:`~pasmopy.construction.reaction_rules.is_dephosphorylated`
          - *pA* is dephosphorylated --> *uA*
          - .. math:: V, K
        * - :func:`~pasmopy.construction.reaction_rules.phosphorylate`
          - *B* phosphorylates *uA* --> *pA*
          - .. math:: V, K
        * - :func:`~pasmopy.construction.reaction_rules.dephosphorylate`
          - *B* dephosphorylates *pA* --> *uA*
          - .. math:: V, K
        * - :func:`~pasmopy.construction.reaction_rules.transcribe`
          - *B* transcribes *a*
          - .. math:: V, K, n, (KF, nF)
        * - :func:`~pasmopy.construction.reaction_rules.is_translated`
          - *a* is translated into *A*
          - .. math:: kf
        * - :func:`~pasmopy.construction.reaction_rules.synthesize`
          - *B* synthesizes *A*
          - .. math:: kf
        * - :func:`~pasmopy.construction.reaction_rules.is_synthesized`
          - *A* is synthesized
          - .. math:: kf
        * - :func:`~pasmopy.construction.reaction_rules.degrade`
          - *B* degrades *A*
          - .. math:: kf
        * - :func:`~pasmopy.construction.reaction_rules.is_degraded`
          - *A* is degraded
          - .. math:: kf
        * - :func:`~pasmopy.construction.reaction_rules.translocate`
          - *Acyt* translocates from cytoplasm to nucleus (Vcyt, Vnuc) <--> *Anuc*
          - .. math:: kf, kr, (V_{pre}, V_{post})

    Attributes
    ----------
    input_txt : str
        Model description file (*.txt), e.g.,
        `Kholodenko1999.txt <https://github.com/pasmopy/pasmopy/blob/master/tests/text_files/Kholodenko1999.txt>`_.
    parameters : list of strings
        ``x`` : model parameters.
    species : list of strings
        ``y`` : model species.
    reactions : list of strings
        ``v`` : flux vector.
    differential_equations : list of strings
        ``dydt`` : right-hand side of the differential equation.
    obs_desc : list of List[str]
        Description of observables.
    param_info : list of strings
        Information about parameter values.
    init_info : list of strings
        Information about initial values.
    param_constraints : list of strings
        Information about parameter constraints.
    param_excluded : list of strings
        List of parameters excluded from search params because of parameter constraints.
    fixed_species : list of strings
        List of species which should be held fixed (never consumed) during simulation.
    sim_tspan : list of strings ['t0', 'tf']
        Interval of integration.
    sim_conditions : list of List[str]
        Simulation conditions with stimulation.
    sim_unperturbed : str
        Untreated conditions to get steady state.
    rule_words : dict
        Words to identify reaction rules.
    complex_formations : list
        List of ComplexFormation to detect duplicate binding-dissociation.

    """

    input_txt: str

    parameters: List[str] = field(
        default_factory=list,
        init=False,
    )
    species: List[str] = field(
        default_factory=list,
        init=False,
    )
    reactions: List[str] = field(
        default_factory=list,
        init=False,
    )
    differential_equations: List[str] = field(
        default_factory=list,
        init=False,
    )
    obs_desc: List[List[str]] = field(
        default_factory=list,
        init=False,
    )
    param_info: List[str] = field(
        default_factory=list,
        init=False,
    )
    init_info: List[str] = field(
        default_factory=list,
        init=False,
    )
    param_constraints: List[str] = field(
        default_factory=list,
        init=False,
    )
    param_excluded: List[str] = field(
        default_factory=list,
        init=False,
    )
    fixed_species: List[str] = field(
        default_factory=list,
        init=False,
    )
    # Information about simulation
    sim_tspan: List[str] = field(
        default_factory=list,
        init=False,
    )
    sim_conditions: List[List[str]] = field(
        default_factory=list,
        init=False,
    )
    sim_unperturbed: str = field(
        default_factory=str,
        init=False,
    )
    # Words to identify reaction rules
    rule_words: Dict[str, List[str]] = field(
        default_factory=lambda: dict(
            dimerize=[
                " dimerizes",
                " homodimerizes",
                " forms a dimer",
                " forms dimers",
            ],
            bind=[
                " binds",
                " forms complexes with",
            ],
            dissociate=[
                " is dissociated into",
                " dissociates to",
            ],
            is_phosphorylated=[
                " is phosphorylated",
            ],
            is_dephosphorylated=[
                " is dephosphorylated",
            ],
            phosphorylate=[
                " phosphorylates",
            ],
            dephosphorylate=[
                " dephosphorylates",
            ],
            transcribe=[
                " transcribe",
                " transcribes",
            ],
            is_translated=[
                " is translated into",
            ],
            synthesize=[
                " synthesizes",
                " promotes synthesis of",
                " increases",
            ],
            is_synthesized=[
                " is synthesized",
            ],
            degrade=[
                " degrades",
                " promotes degradation of",
                " decreases",
            ],
            is_degraded=[
                " is degraded",
            ],
            translocate=[
                " translocates",
                " is translocated",
            ],
        ),
        init=False,
    )

    @staticmethod
    def _isfloat(string: str) -> bool:
        """
        Checking if a string can be converted to float.
        """
        try:
            float(string)
            return True
        except ValueError:
            return False

    @staticmethod
    def _remove_prefix(text: str, prefix: str) -> str:
        """
        Remove prefix from a text.
        """
        if text.startswith(prefix):
            return text[len(prefix) :]
        assert False

    def _set_params(self, line_num: int, *args: str) -> None:
        """
        Set model parameters.
        """
        for p_name in args:
            if p_name + f"{line_num:d}" not in self.parameters:
                self.parameters.append(p_name + f"{line_num:d}")

    def _set_species(self, *args: str) -> None:
        """
        Set model species.
        """
        for s_name in args:
            if s_name not in self.species:
                self.species.append(s_name)

    def _preprocessing(
        self,
        func_name: str,
        line_num: int,
        line: str,
        *args: str,
    ) -> List[str]:
        """
        Extract the information about parameter and/or initial values
        if '|' in the line and find a keyword to identify reaction rules.

        Parameters
        ----------
        func_name : str
            Name of the rule function.

        line_num : int
            Line number.

        line : str
            Each line of the input text.

        Returns
        -------
        description : list of strings

        """
        self._set_params(line_num, *args)
        if "|" in line:
            if line.split("|")[1].strip():
                param_values = line.split("|")[1].strip().split(",")
                if all("=" in pval for pval in param_values):
                    for pval in param_values:
                        base_param = pval.split("=")[0].strip(" ")
                        if base_param.startswith("const "):
                            # Parameter names with 'const' will be added to param_excluded.
                            base_param = base_param.split("const ")[-1]
                            fixed = True
                        else:
                            fixed = False
                        if base_param in args:
                            if self._isfloat(pval.split("=")[1].strip(" ")):
                                self.param_info.append(
                                    "x[C."
                                    + base_param
                                    + f"{line_num:d}] = "
                                    + pval.split("=")[1].strip(" ")
                                )
                                # If a parameter value is initialized to 0.0 or fixed,
                                # then add it to param_excluded.
                                if float(pval.split("=")[1].strip(" ")) == 0.0 or fixed:
                                    self.param_excluded.append(base_param + f"{line_num:d}")
                            else:
                                raise ValueError(
                                    f"line{line_num:d}: Parameter value must be int or float."
                                )
                        else:
                            raise ValueError(
                                f"line{line_num:d}: '{pval.split('=')[0].strip(' ')}'\n"
                                f"Available parameters are: {', '.join(args)}."
                            )
                elif param_values[0].strip(" ").isdecimal():
                    # Parameter constraints
                    for param_name in args:
                        if f"{param_name}{int(param_values[0]):d}" not in self.parameters:
                            raise ValueError(
                                f"Line {line_num:d} and {int(param_values[0]):d} : "
                                "Different reaction rules in parameter constraints."
                            )
                        else:
                            self.param_excluded.append(f"{param_name}{line_num:d}")
                            self.param_info.append(
                                f"x[C.{param_name}"
                                f"{line_num:d}] = "
                                f"x[C.{param_name}"
                                f"{int(param_values[0]):d}]"
                            )
                            self.param_constraints.append(
                                f"x[C.{param_name}"
                                f"{line_num:d}] = "
                                f"x[C.{param_name}"
                                f"{int(param_values[0]):d}]"
                            )
                else:
                    raise ValueError(
                        f"line{line_num:d}: {line}\nInvalid expression in the input parameter."
                    )
            if line.count("|") > 1 and line.split("|")[2].strip():
                initial_values = line.split("|")[2].strip().split(",")
                for ival in initial_values:
                    if ival.startswith("fixed "):
                        ival = ival.split("fixed ")[-1]
                        self.fixed_species.append(ival.split("=")[0].strip(" "))
                    if ival.split("=")[0].strip(" ") in line.split("|")[0]:
                        if self._isfloat(ival.split("=")[1].strip(" ")):
                            self.init_info.append(
                                "y0[V."
                                + ival.split("=")[0].strip(" ")
                                + "] = "
                                + ival.split("=")[1].strip(" ")
                            )
                        else:
                            raise ValueError(
                                f"line{line_num:d}: Initial value must be int or float."
                            )
                    else:
                        raise NameError(
                            f"line{line_num:d}: "
                            f"Name'{ival.split('=')[0].strip(' ')}' is not defined."
                        )
            line = line.split("|")[0]
        hit_words: List[str] = []
        for word in self.rule_words[func_name]:
            # Choose longer word
            if word in line:
                hit_words.append(word)

        return line.strip().split(max(hit_words, key=len))

    @staticmethod
    def _word2scores(word: str, sentence: str) -> List[float]:
        """
        Calculate similarity scores between word and sentence.

        Parameters
        ----------
        word : str
            User-defined word.
        sentence : str
            Textual unit consisting of two or more words.

        returns
        -------
        ratio : list
            List containing similarity scores.

        """
        ratio = [
            SequenceMatcher(None, word, sentence[i : i + len(word)]).ratio()
            for i in range(len(sentence) - len(word) + 1)
        ]
        return ratio

    def _get_partial_similarity(
        self,
        line: str,
        similarity_threshold: float = 0.7,
    ) -> UnregisteredRule:
        """
        Suggest similar rule word when user-defined word is not registered
        in rule_words.

        Parameters
        ----------
        line : str
            Each line of the input text.

        similarity_threshold : float (default: 0.7)
            if all match_scores are below this value, expected_word will not
            be returned.

        Returns
        -------
        unregistered_rule : UnregisteredRule
            Rule word with the highest similarity score.

        """
        match_words = []
        match_scores = []
        str_subset = []
        for rules in self.rule_words.values():
            for word in rules:
                ratio = self._word2scores(word, line)
                if ratio:
                    match_words.append(word)
                    match_scores.append(max(ratio))
                    str_subset.append(line[np.argmax(ratio) : np.argmax(ratio) + len(word)])
        expected_word = (
            None
            if all([score < similarity_threshold for score in match_scores])
            else match_words[np.argmax(match_scores)]
        )
        original_word = (
            None if expected_word is None else str_subset[match_words.index(expected_word)]
        )
        unregistered_rule = UnregisteredRule(expected_word, original_word)

        return unregistered_rule

    @staticmethod
    def _remove_prepositions(sentence: str) -> str:
        """
        Remove preposition from text not to use it for identifying reaction rules.
        """
        for preposition in PREPOSITIONS:
            if sentence.endswith(f" {preposition}"):
                return sentence[: -len(preposition) - 1]
        return sentence

    def dimerize(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'A dimerizes <--> AA'
        >>> 'A homodimerizes <--> AA'
        >>> 'A forms a dimer <--> AA'
        >>> 'A forms dimers <--> AA'

        Notes
        -----
        * Parameters
            .. math:: kf, kr

        * Rate equation
            .. math:: v = kf * [A] * [A] - kr * [AA]

        * Differential equation
            .. math::

                d[A]]/dt = - 2 * v

                d[AA]/dt = + v

        """
        description = self._preprocessing(
            sys._getframe().f_code.co_name, line_num, line, "kf", "kr"
        )
        monomer = description[0].strip(" ")
        if " <--> " in description[1]:
            dimer = description[1].split(" <--> ")[1].strip(" ")
        elif " --> " in description[1]:
            warnings.warn(
                f"line{line_num:d}: Use '<-->' instead of '-->' for reversible reaction rules.",
                FutureWarning,
            )
            dimer = description[1].split(" --> ")[1].strip(" ")
        else:
            raise ValueError(f"line{line_num:d}: Use '<-->' to specify the name of the dimer.")
        if monomer == dimer:
            raise ValueError(f"{dimer} <- Use a different name.")
        self._set_species(monomer, dimer)
        self.complex_formations.append(ComplexFormation(line_num, set(monomer), dimer, True))
        self.reactions.append(
            f"v[{line_num:d}] = "
            f"x[C.kf{line_num:d}] * y[V.{monomer}] * y[V.{monomer}] - "
            f"x[C.kr{line_num:d}] * y[V.{dimer}]"
        )
        counter_monomer, counter_dimer = (0, 0)
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{monomer}]" in eq:
                counter_monomer += 1
                self.differential_equations[i] = eq + f" - 2 * v[{line_num:d}]"
            elif f"dydt[V.{dimer}]" in eq:
                counter_dimer += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
        if counter_monomer == 0:
            self.differential_equations.append(f"dydt[V.{monomer}] = - v[{line_num:d}]")
        if counter_dimer == 0:
            self.differential_equations.append(f"dydt[V.{dimer}] = + v[{line_num:d}]")

    def bind(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'A binds B <--> AB'
        >>> 'A forms complexes with B <--> AB'

        Notes
        -----
        * Parameters
            .. math:: kf, kr

        * Rate equation
            .. math:: v = kf * [A] * [B] - kr * [AB]

        * Differential equation
            .. math::

                d[A]/dt = - v

                d[B]/dt = - v

                d[AB]/dt = + v

        """
        description = self._preprocessing(
            sys._getframe().f_code.co_name, line_num, line, "kf", "kr"
        )
        component1 = description[0].strip(" ")
        if " <--> " in description[1]:
            # Specify name of the complex
            component2 = description[1].split(" <--> ")[0].strip(" ")
            complex = description[1].split(" <--> ")[1].strip(" ")
        elif " --> " in description[1]:
            warnings.warn(
                f"line{line_num:d}: Use '<-->' instead of '-->' for reversible reaction rules.",
                FutureWarning,
            )
            component2 = description[1].split(" --> ")[0].strip(" ")
            complex = description[1].split(" --> ")[1].strip(" ")
        else:
            raise ValueError(
                f"line{line_num:d}: Use '<-->' to specify the name of the protein complex."
            )
        if component1 == complex or component2 == complex:
            raise ValueError(f"line{line_num:d}: {complex} <- Use a different name.")
        elif component1 == component2:
            self.dimerize(line_num, line)
        else:
            self._set_species(component1, component2, complex)
            self.complex_formations.append(
                ComplexFormation(line_num, set([component1, component2]), complex, True)
            )
            self.reactions.append(
                f"v[{line_num:d}] = "
                f"x[C.kf{line_num:d}] * y[V.{component1}] * y[V.{component2}] - "
                f"x[C.kr{line_num:d}] * y[V.{complex}]"
            )
            counter_component1, counter_component2, counter_complex = (0, 0, 0)
            for i, eq in enumerate(self.differential_equations):
                if f"dydt[V.{component1}]" in eq:
                    counter_component1 += 1
                    self.differential_equations[i] = eq + f" - v[{line_num:d}]"
                elif f"dydt[V.{component2}]" in eq:
                    counter_component2 += 1
                    self.differential_equations[i] = eq + f" - v[{line_num:d}]"
                elif f"dydt[V.{complex}]" in eq:
                    counter_complex += 1
                    self.differential_equations[i] = eq + f" + v[{line_num:d}]"
            if counter_component1 == 0:
                self.differential_equations.append(f"dydt[V.{component1}] = - v[{line_num:d}]")
            if counter_component2 == 0:
                self.differential_equations.append(f"dydt[V.{component2}] = - v[{line_num:d}]")
            if counter_complex == 0:
                self.differential_equations.append(f"dydt[V.{complex}] = + v[{line_num:d}]")

    def dissociate(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'AB dissociates to A and B'
        >>> 'AB is dissociated into A and B'

        Notes
        -----
        * Parameters
            .. math:: kf, kr

        * Rate equation
            .. math:: v = kf * [AB] - kr * [A] * [B]

        * Differential equation
            .. math::

                d[A]/dt = + v

                d[B]/dt = + v

                d[AB]/dt = - v

        """
        description = self._preprocessing(
            sys._getframe().f_code.co_name, line_num, line, "kf", "kr"
        )
        complex = description[0].strip(" ")
        if " and " not in description[1]:
            raise ValueError(
                f"Use 'and' in line{line_num:d}:\ne.g., AB is dissociated into A and B"
            )
        else:
            component1 = description[1].split(" and ")[0].strip(" ")
            component2 = description[1].split(" and ")[1].strip(" ")
        self._set_species(complex, component1, component2)
        self.complex_formations.append(
            ComplexFormation(line_num, set([component1, component2]), complex, False)
        )
        self.reactions.append(
            f"v[{line_num:d}] = "
            f"x[C.kf{line_num:d}] * y[V.{complex}] - "
            f"x[C.kr{line_num:d}] * y[V.{component1}] * y[V.{component2}]"
        )
        counter_complex, counter_component1, counter_component2 = (0, 0, 0)
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{complex}]" in eq:
                counter_complex += 1
                self.differential_equations[i] = eq + f" - v[{line_num:d}]"
            elif f"dydt[V.{component1}]" in eq:
                counter_component1 += 1
                self.differential_equations[i] = (
                    eq + f" + v[{line_num:d}]"
                    if component1 != component2
                    else eq + f" + 2 * v[{line_num:d}]"
                )
            elif f"dydt[V.{component2}]" in eq:
                counter_component2 += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
        if counter_complex == 0:
            self.differential_equations.append(f"dydt[V.{complex}] = - v[{line_num:d}]")
        if counter_component1 == 0:
            self.differential_equations.append(f"dydt[V.{component1}] = + v[{line_num:d}]")
        if counter_component2 == 0:
            self.differential_equations.append(f"dydt[V.{component2}] = + v[{line_num:d}]")

    def is_phosphorylated(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'uA is phosphorylated <--> pA'

        Notes
        -----
        * Parameters
            .. math:: kf, kr

        * Rate equation
            .. math:: v = kf * [uA] - kr * [pA]

        * Differential equation
            .. math::

                d[uA]/dt = - v

                d[pA]/dt = + v

        """
        description = self._preprocessing(
            sys._getframe().f_code.co_name, line_num, line, "kf", "kr"
        )
        unphosphorylated_form = description[0].strip(" ")
        if " <--> " in description[1]:
            phosphorylated_form = description[1].split(" <--> ")[1].strip(" ")
        elif " --> " in description[1]:
            warnings.warn(
                f"line{line_num:d}: Use '<-->' instead of '-->' for reversible reaction rules.",
                FutureWarning,
            )
            phosphorylated_form = description[1].split(" --> ")[1].strip(" ")
        else:
            raise ValueError(
                f"line{line_num:d}: "
                "Use '<-->' to specify the name of the phosphorylated protein."
            )
        self._set_species(unphosphorylated_form, phosphorylated_form)

        self.reactions.append(
            f"v[{line_num:d}] = "
            f"x[C.kf{line_num:d}] * y[V.{unphosphorylated_form}] - "
            f"x[C.kr{line_num:d}] * y[V.{phosphorylated_form}]"
        )
        counter_unphosphorylated_form, counter_phosphorylated_form = (0, 0)
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{unphosphorylated_form}]" in eq:
                counter_unphosphorylated_form += 1
                self.differential_equations[i] = eq + f" - v[{line_num:d}]"
            elif "dydt[V.{phosphorylated_form}]" in eq:
                counter_phosphorylated_form += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
        if counter_unphosphorylated_form == 0:
            self.differential_equations.append(
                f"dydt[V.{unphosphorylated_form}] = - v[{line_num:d}]"
            )
        if counter_phosphorylated_form == 0:
            self.differential_equations.append(
                f"dydt[V.{phosphorylated_form}] = + v[{line_num:d}]"
            )

    def is_dephosphorylated(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'pA is dephosphorylated --> uA'

        Notes
        -----
        * Parameters
            .. math:: V, K

        * Rate equation
            .. math:: v = V * [pA] / (K + [pA])

        * Differential equation
            .. math::

                d[uA]/dt = + v

                d[pA]/dt = - v

        """
        description = self._preprocessing(sys._getframe().f_code.co_name, line_num, line, "V", "K")
        phosphorylated_form = description[0].strip(" ")
        if " --> " in description[1]:
            unphosphorylated_form = description[1].split(" --> ")[1].strip(" ")
        else:
            raise ValueError(
                f"line{line_num:d}: "
                "Use '-->' to specify the name of the dephosphorylated protein."
            )
        self._set_species(phosphorylated_form, unphosphorylated_form)

        self.reactions.append(
            f"v[{line_num:d}] = "
            f"x[C.V{line_num:d}] * y[V.{phosphorylated_form}] / "
            f"(x[C.K{line_num:d}] + y[V.{phosphorylated_form}])"
        )
        counter_unphosphorylated_form, counter_phosphorylated_form = (0, 0)
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{unphosphorylated_form}]" in eq:
                counter_unphosphorylated_form += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
            elif f"dydt[V.{phosphorylated_form}]" in eq:
                counter_phosphorylated_form += 1
                self.differential_equations[i] = eq + f" - v[{line_num:d}]"
        if counter_unphosphorylated_form == 0:
            self.differential_equations.append(
                f"dydt[V.{unphosphorylated_form}] = + v[{line_num:d}]"
            )
        if counter_phosphorylated_form == 0:
            self.differential_equations.append(
                f"dydt[V.{phosphorylated_form}] = - v[{line_num:d}]"
            )

    def phosphorylate(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'B phosphorylates uA --> pA'

        Notes
        -----
        * Parameters
            .. math:: V, K

        * Rate equation
            .. math:: v = V * [B] * [uA] / (K + [uA])

        * Differential equation
            .. math::

                d[uA]/dt = - v

                d[pA]/dt = + v

        """
        description = self._preprocessing(sys._getframe().f_code.co_name, line_num, line, "V", "K")
        kinase = description[0].strip(" ")
        if " --> " in description[1]:
            unphosphorylated_form = description[1].split(" --> ")[0].strip(" ")
            phosphorylated_form = description[1].split(" --> ")[1].strip(" ")
        else:
            raise ValueError(
                f"line{line_num:d}: "
                "Use '-->' to specify the name of the phosphorylated "
                "(or activated) protein."
            )
        if unphosphorylated_form == phosphorylated_form:
            raise ValueError(f"line{line_num:d}: {phosphorylated_form} <- Use a different name.")
        self._set_species(kinase, unphosphorylated_form, phosphorylated_form)

        self.reactions.append(
            f"v[{line_num:d}] = "
            f"x[C.V{line_num:d}] * y[V.{kinase}] * y[V.{unphosphorylated_form}] / "
            f"(x[C.K{line_num:d}] + y[V.{unphosphorylated_form}])"
        )
        counter_unphosphorylated_form, counter_phosphorylated_form = (0, 0)
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{unphosphorylated_form}]" in eq:
                counter_unphosphorylated_form += 1
                self.differential_equations[i] = eq + f" - v[{line_num:d}]"
            elif f"dydt[V.{phosphorylated_form}]" in eq:
                counter_phosphorylated_form += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
        if counter_unphosphorylated_form == 0:
            self.differential_equations.append(
                f"dydt[V.{unphosphorylated_form}] = - v[{line_num:d}]"
            )
        if counter_phosphorylated_form == 0:
            self.differential_equations.append(
                f"dydt[V.{phosphorylated_form}] = + v[{line_num:d}]"
            )

    def dephosphorylate(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'B dephosphorylates pA --> uA'

        Notes
        -----
        * Parameters
            .. math:: V, K

        * Rate equation
            .. math:: v = V * [B] * [pA] / (K + [pA])

        * Differential equation
            .. math::

                d[uA]/dt = + v

                d[pA]/dt = - v

        """
        description = self._preprocessing(sys._getframe().f_code.co_name, line_num, line, "V", "K")
        phosphatase = description[0].strip(" ")
        if " --> " in description[1]:
            phosphorylated_form = description[1].split(" --> ")[0].strip(" ")
            unphosphorylated_form = description[1].split(" --> ")[1].strip(" ")
        else:
            raise ValueError(
                f"line{line_num:d}: "
                "Use '-->' to specify the name of the dephosphorylated "
                "(or deactivated) protein."
            )
        if phosphorylated_form == unphosphorylated_form:
            raise ValueError(f"line{line_num:d}: {unphosphorylated_form} <- Use a different name.")
        self._set_species(phosphatase, phosphorylated_form, unphosphorylated_form)

        self.reactions.append(
            f"v[{line_num:d}] = "
            f"x[C.V{line_num:d}] * y[V.{phosphatase}] * y[V.{phosphorylated_form}] / "
            f"(x[C.K{line_num:d}] + y[V.{phosphorylated_form}])"
        )
        counter_phosphorylated_form, counter_unphosphorylated_form = (0, 0)
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{phosphorylated_form}]" in eq:
                counter_phosphorylated_form += 1
                self.differential_equations[i] = eq + f" - v[{line_num:d}]"
            elif f"dydt[V.{unphosphorylated_form}]" in eq:
                counter_unphosphorylated_form += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
        if counter_phosphorylated_form == 0:
            self.differential_equations.append(
                f"dydt[V.{phosphorylated_form}] = - v[{line_num:d}]"
            )
        if counter_unphosphorylated_form == 0:
            self.differential_equations.append(
                f"dydt[V.{unphosphorylated_form}] = + v[{line_num:d}]"
            )

    def transcribe(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'B transcribes a'
        >>> 'B1 & B2 transcribe a'  # (AND-gate)
        >>> 'B transcribes a, repressed by C'  # (Negative regulation)

        Notes
        -----
        * Parameters
            .. math:: V, K, n, (KF, nF)

        * Rate equation
            .. math::

                v = V * [B] ^ {n} / (K ^ {n} + [B] ^ {n})

                v = V * ([B1] * [B2]) ^ {n} / (K ^ {n} + ([B1] * [B2]) ^ {n})

                v = V * [B] ^ {n} / (K ^ {n} + [B] ^ {n} + ([C] / KF) ^ {nF})

        * Differential equation
            .. math:: d[a]/dt = + v

        """
        description = self._preprocessing(
            sys._getframe().f_code.co_name, line_num, line, "V", "K", "n", "KF", "nF"
        )
        repressor: Optional[str] = None
        ratio = self._word2scores(", repressed by", description[1])
        if not ratio or max(ratio) < 1.0:
            self.parameters.remove(f"KF{line_num:d}")
            self.parameters.remove(f"nF{line_num:d}")
            mRNA = description[1].strip()
            if " " in mRNA:
                # Fix typo in line{line_num:d}
                raise ValueError(
                    f"line{line_num:d}: "
                    "Add ', repressed by XXX' to describe negative regulation from XXX."
                )
        else:
            # Add negative regulation from repressor
            mRNA = description[1].split(", repressed by")[0].strip()
            repressor = description[1].split(", repressed by")[1].strip()
        if " & " not in description[0]:
            TF = description[0].strip(" ")
            self._set_species(mRNA, TF)
            if repressor is not None:
                self._set_species(repressor)
            self.reactions.append(
                f"v[{line_num:d}] = "
                f"x[C.V{line_num:d}] * y[V.{TF}] ** x[C.n{line_num:d}] / "
                f"(x[C.K{line_num:d}] ** x[C.n{line_num:d}] + "
                f"y[V.{TF}] ** x[C.n{line_num:d}]"
                + (
                    ")"
                    if repressor is None
                    else f" + (y[V.{repressor}] / x[C.KF{line_num:d}]) ** x[C.nF{line_num:d}])"
                )
            )
        else:
            # AND-gate
            TFs = [TF.strip(" ") for TF in description[0].split(" & ")]
            self._set_species(mRNA, *TFs)
            if repressor is not None:
                self._set_species(repressor)
            self.reactions.append(
                f"v[{line_num:d}] = "
                f"x[C.V{line_num:d}] * ({'y[V.' + '] * y[V.'.join(TFs) + ']'}) ** x[C.n{line_num:d}] / "
                f"(x[C.K{line_num:d}] ** x[C.n{line_num:d}] + "
                f"({'y[V.' + '] * y[V.'.join(TFs) + ']'}) ** x[C.n{line_num:d}]"
                + (
                    ")"
                    if repressor is None
                    else f" + (y[V.{repressor}] / x[C.KF{line_num:d}]) ** x[C.nF{line_num:d}])"
                )
            )
        counter_mRNA = 0
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{mRNA}]" in eq:
                counter_mRNA += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
        if counter_mRNA == 0:
            self.differential_equations.append(f"dydt[V.{mRNA}] = + v[{line_num:d}]")

    def is_translated(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'a is translated into A'

        Notes
        -----
        * Parameters
            .. math:: kf

        * Rate equation
            .. math:: v = kf * [a]

        * Differential equation
            .. math:: d[A]/dt = + v

        """
        description = self._preprocessing(sys._getframe().f_code.co_name, line_num, line, "kf")
        mRNA = description[0].strip(" ")
        protein = description[1].strip(" ")
        self._set_species(mRNA, protein)
        self.reactions.append(f"v[{line_num:d}] = x[C.kf{line_num:d}] * y[V.{mRNA}]")
        counter_protein = 0
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{protein}]" in eq:
                counter_protein += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
        if counter_protein == 0:
            self.differential_equations.append(f"dydt[V.{protein}] = + v[{line_num:d}]")

    def synthesize(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'B synthesizes A'

        Notes
        -----
        * Parameters
            .. math:: kf

        * Rate equation
            .. math:: v = kf * [B]

        * Differential equation
            .. math:: d[A]/dt = + v

        """
        description = self._preprocessing(sys._getframe().f_code.co_name, line_num, line, "kf")
        catalyst = description[0].strip(" ")
        product = description[1].strip(" ")
        self._set_species(catalyst, product)
        self.reactions.append(f"v[{line_num:d}] = x[C.kf{line_num:d}] * y[V.{catalyst}]")
        counter_product = 0
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{product}]" in eq:
                counter_product += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
        if counter_product == 0:
            self.differential_equations.append(f"dydt[V.{product}] = + v[{line_num:d}]")

    def is_synthesized(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'A is synthesized'

        Notes
        -----
        * Parameters
            .. math:: kf

        * Rate equation
            .. math:: v = kf

        * Differential equation
            .. math:: d[A]/dt = + v

        """
        description = self._preprocessing(sys._getframe().f_code.co_name, line_num, line, "kf")
        chemical_species = description[0].strip(" ")
        self._set_species(chemical_species)
        self.reactions.append(f"v[{line_num:d}] = x[C.kf{line_num:d}]")
        counter_chemical_species = 0
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{chemical_species}]" in eq:
                counter_chemical_species += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
        if counter_chemical_species == 0:
            self.differential_equations.append(f"dydt[V.{chemical_species}] = + v[{line_num:d}]")

    def degrade(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'B degrades A'

        Notes
        -----
        * Parameters
            .. math:: kf

        * Rate equation
            .. math:: v = kf * [B]

        * Differential equation
            .. math:: d[A]/dt = - v

        """
        description = self._preprocessing(sys._getframe().f_code.co_name, line_num, line, "kf")
        protease = description[0].strip(" ")
        protein = description[1].strip(" ")
        self._set_species(protease, protein)
        self.reactions.append(f"v[{line_num:d}] = x[C.kf{line_num:d}] * y[V.{protease}]")
        counter_protein = 0
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{protein}]" in eq:
                counter_protein += 1
                self.differential_equations[i] = eq + f" - v[{line_num:d}]"
        if counter_protein == 0:
            self.differential_equations.append(f"dydt[V.{protein}] = - v[{line_num:d}]")

    def is_degraded(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'A is degraded'

        Notes
        -----
        * Parameters
            .. math:: kf

        * Rate equation
            .. math:: v = kf * [A]

        * Differential equation
            .. math:: d[A]/dt = - v

        """
        description = self._preprocessing(sys._getframe().f_code.co_name, line_num, line, "kf")
        chemical_species = description[0].strip(" ")
        self._set_species(chemical_species)
        self.reactions.append(f"v[{line_num:d}] = x[C.kf{line_num:d}] * y[V.{chemical_species}]")
        counter_chemical_species = 0
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{chemical_species}]" in eq:
                counter_chemical_species += 1
                self.differential_equations[i] = eq + f" - v[{line_num:d}]"
        if counter_chemical_species == 0:
            self.differential_equations.append(f"dydt[V.{chemical_species}] = - v[{line_num:d}]")

    def translocate(self, line_num: int, line: str) -> None:
        r"""
        Examples
        --------
        >>> 'A_at_cyt translocates from cytoplasm to nucleus (V_cyt, V_nuc) <--> A_at_nuc'
        >>> 'A_at_cyt is translocated from cytoplasm to nucleus (V_cyt, V_nuc) <--> A_at_nuc'

        Notes
        -----
        * Parameters
            .. math:: kf, kr, (V_{pre}, V_{post})

        * Rate equation
            .. math:: v = kf * [A\_at\_pre] - kr * (V_{post} / V_{pre}) * [A\_at\_post]

        * Differential equation
            .. math::

                d[A\_at\_pre]/dt = - v

                d[A\_at\_post]/dt = + v * (V_{pre} / V_{post})

        """
        description = self._preprocessing(
            sys._getframe().f_code.co_name, line_num, line, "kf", "kr"
        )
        pre_translocation = description[0].strip(" ")
        if " <--> " in description[1]:
            post_translocation = description[1].split(" <--> ")[1].strip(" ")
        elif " --> " in description[1]:
            warnings.warn(
                f"line{line_num:d}: Use '<-->' instead of '-->' for reversible reaction rules.",
                FutureWarning,
            )
            post_translocation = description[1].split(" --> ")[1].strip(" ")
        else:
            raise ValueError(
                f"line{line_num:d}: "
                "Use '<-->' to specify the name of the species after translocation."
            )
        if pre_translocation == post_translocation:
            raise ValueError(f"line{line_num:d}: {post_translocation} <- Use a different name.")
        # Information about compartment volumes
        if "(" in description[1] and ")" in description[1]:
            [pre_volume, post_volume] = description[1].split("(")[-1].split(")")[0].split(",")
            if not self._isfloat(pre_volume.strip(" ")) or not self._isfloat(
                post_volume.strip(" ")
            ):
                raise ValueError("pre_volume and post_volume must be float or int.")
        else:
            [pre_volume, post_volume] = ["1", "1"]
        self._set_species(pre_translocation, post_translocation)
        self.reactions.append(
            f"v[{line_num:d}] = x[C.kf{line_num:d}] * y[V.{pre_translocation}] - "
            f"x[C.kr{line_num:d}] * y[V.{post_translocation}]"
        )
        if float(pre_volume.strip(" ")) != float(post_volume.strip(" ")):
            self.reactions[-1] = (
                f"v[{line_num:d}] = "
                f"x[C.kf{line_num:d}] * y[V.{pre_translocation}] - "
                f"x[C.kr{line_num:d}] * "
                f"({post_volume.strip()} / {pre_volume.strip()}) * "
                f"y[V.{post_translocation}]"
            )
        counter_pre_translocation, counter_post_translocation = (0, 0)
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{pre_translocation}]" in eq:
                counter_pre_translocation += 1
                self.differential_equations[i] = eq + f" - v[{line_num:d}]"
            elif f"dydt[V.{post_translocation}]" in eq:
                counter_post_translocation += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
                if float(pre_volume.strip(" ")) != float(post_volume.strip(" ")):
                    self.differential_equations[
                        i
                    ] += f" * ({pre_volume.strip()} / {post_volume.strip()})"
        if counter_pre_translocation == 0:
            self.differential_equations.append(f"dydt[V.{pre_translocation}] = - v[{line_num:d}]")
        if counter_post_translocation == 0:
            self.differential_equations.append(f"dydt[V.{post_translocation}] = + v[{line_num:d}]")
            if float(pre_volume.strip(" ")) != float(post_volume.strip(" ")):
                self.differential_equations[
                    -1
                ] += f" * ({pre_volume.strip()} / {post_volume.strip()})"

    def create_ode(self) -> None:
        """
        Find a keyword in each line to identify the reaction rule and
        construct an ODE model.

        """
        with open(self.input_txt, encoding="utf-8") as f:
            lines = f.readlines()
        for line_num, line in enumerate(lines, start=1):
            # Remove double spaces
            while True:
                if "  " not in line:
                    break
                else:
                    line = line.replace("  ", " ")
            # Comment out
            line = line.split("#")[0].rstrip(" ")
            if not line.strip():
                # Skip blank lines
                continue
            elif lines.count(line) > 1:
                # Find duplicate lines
                raise DuplicateError(
                    f"Reaction '{line}' is duplicated in lines "
                    + ", ".join([str(i + 1) for i, rxn in enumerate(lines) if rxn == line])
                )
            # About observables
            elif line.startswith("@obs "):
                line = self._remove_prefix(line, "@obs ")
                if line.count(":") != 1:
                    raise SyntaxError(
                        f"line{line_num:d}: Missing colon\n"
                        "Should be `@obs <observable name>: <expression>`."
                    )
                else:
                    self.obs_desc.append(line.split(":"))
            # About simulation info.
            elif line.startswith("@sim "):
                line = self._remove_prefix(line, "@sim ")
                if line.count(":") != 1:
                    raise SyntaxError(f"line{line_num:d}: Missing colon")
                else:
                    if line.startswith("tspan"):
                        t_info = line.split(":")[-1].strip()
                        if "[" in t_info and "]" in t_info:
                            [t0, tf] = t_info.split("[")[-1].split("]")[0].split(",")
                            if t0.strip(" ").isdecimal() and tf.strip(" ").isdecimal():
                                self.sim_tspan.append(t0)
                                self.sim_tspan.append(tf)
                            else:
                                raise TypeError("@sim tspan: [t0, tf] must be a list of integers.")
                        else:
                            raise ValueError(
                                "`tspan` must be a two element vector [t0, tf] "
                                "specifying the initial and final times."
                            )
                    elif line.startswith("unperturbed"):
                        self.sim_unperturbed += line.split(":")[-1].strip()
                    elif line.startswith("condition "):
                        self.sim_conditions.append(
                            self._remove_prefix(line, "condition ").split(":")
                        )
                    else:
                        raise ValueError(
                            f"(line{line_num:d}) Available options are: "
                            "'@sim tspan:', '@sim unperturbed:', or '@sim condition XXX:'."
                        )
            # Detect reaction rule
            else:
                for reaction_rule, words in self.rule_words.items():
                    if any([self._remove_prepositions(word) in line for word in words]):
                        exec("self." + reaction_rule + "(line_num, line)")
                        break
                else:
                    unregistered_rule = self._get_partial_similarity(line)
                    raise ValueError(
                        f"Unregistered words in line{line_num:d}: {line}"
                        + (
                            f"\nMaybe: '{unregistered_rule.expected}'."
                            if unregistered_rule.expected is not None
                            else ""
                        )
                    )



import sys
from pathlib import Path

from setuptools import find_packages, setup

try:
    from pasmopy import __author__, __email__, __maintainer__
except ImportError:
    __author__ = __maintainer__ = "Hiroaki Imoto"
    __email__ = "himoto@protein.osaka-u.ac.jp"


def get_version() -> str:
    """Read version from file"""
    version_filepath = Path(__file__).parent / "pasmopy" / "version.py"
    with open(version_filepath) as f:
        for line in f:
            if line.startswith("__version__"):
                return line.strip().split()[-1][1:-1]
    assert False


def setup_package():
    # Python version check.
    if sys.version_info[:2] < (3, 7):
        raise RuntimeError("Pasmopy requires at least Python version 3.7")

    setup(
        name="pasmopy",
        version=get_version(),
        description="Patient-Specific Modeling in Python",
        long_description=Path("README.md").read_text("utf-8"),
        long_description_content_type="text/markdown",
        license="Apache 2.0",
        author=__author__,
        author_email=__email__,
        maintainer=__maintainer__,
        maintainer_email=__email__,
        url="https://github.com/pasmopy/pasmopy",
        download_url="https://github.com/pasmopy/pasmopy/releases",
        project_urls={
            "Documentation": "https://pasmopy.readthedocs.io/en/latest/",
            "Source Code": "https://github.com/pasmopy/pasmopy",
            "Bug Tracker": "https://github.com/pasmopy/pasmopy/issues",
        },
        packages=find_packages(exclude=["tests", "docs"]),
        install_requires=[
            l.strip() for l in Path("requirements.txt").read_text("utf-8").splitlines()
        ],
        extras_require={
            "dev": [
                "black==22.1.0",
                "flake8==4.0.1",
                "isort==5.10.1",
                "pre-commit",
                "pytest",
            ],
            "docs": [
                "sphinx>=1.7",
                "sphinx_rtd_theme>=0.3",
                "sphinx_autodoc_typehints>=1.10",
                "sphinxcontrib-bibtex>=2.2",
            ],
        },
        python_requires=">=3.7",
        keywords=[
            "systems",
            "biology",
            "cancer",
            "stratification",
            "personalized",
            "modeling",
            "simulation",
            "precision",
            "oncology",
        ],
        classifiers=[
            "Intended Audience :: Science/Research",
            "Intended Audience :: Healthcare Industry",
            "License :: OSI Approved :: Apache Software License",
            "Natural Language :: English",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Topic :: Scientific/Engineering",
            "Topic :: Scientific/Engineering :: Bio-Informatics",
            "Topic :: Scientific/Engineering :: Medical Science Apps.",
            "Topic :: Software Development",
            "Topic :: Software Development :: Libraries",
            "Topic :: Software Development :: Libraries :: Python Modules",
        ],
        zip_safe=False,
    )


if __name__ == "__main__":
    setup_package()



from dataclasses import make_dataclass
from typing import Dict, List

NAMES: List[str] = []

NUM: int = len(NAMES)

Parameters = make_dataclass(
    cls_name="Parameters",
    fields=[(name, int) for name in NAMES],
    namespace={"NAMES": NAMES, "NUM": NUM},
    frozen=True,
)

name2idx: Dict[str, int] = {k: v for v, k in enumerate(NAMES)}

C = Parameters(**name2idx)

del name2idx
